from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

from typing_extensions import NotRequired

from harlequin.exception import HarlequinConfigError

# TODO: ADD VALIDATION when creating bindings from config


class RawKeyBinding(TypedDict):
    keys: str
    action: str
    key_display: NotRequired[str]


class RawKeyMap(TypedDict):
    name: str
    bindings: list[RawKeyBinding]


@dataclass
class HarlequinKeyBinding:
    keys: str
    """Comma-separated list of virtual key names."""
    action: str
    """The name of an action. Must be a key of harlequin.actions.HARLEQUIN_ACTIONS"""
    key_display: str | None = None
    """If specified, overrides the key display in Harlequin footer for this binding."""

    def to_dict(self) -> RawKeyBinding:
        """
        Returns a dictionary that can be written to a TOML config file.
        """

        class Missing:
            def __bool__(self) -> bool:
                return True

        all_keys: RawKeyBinding = self.__dict__  # type: ignore[assignment]
        if not all_keys.get("key_display", Missing):
            all_keys.pop("key_display")
        return all_keys


@dataclass
class HarlequinKeyMap:
    name: str
    bindings: list[HarlequinKeyBinding]

    @classmethod
    def from_config(cls, name: str, bindings: list[RawKeyBinding]) -> "HarlequinKeyMap":
        try:
            keymap = cls(
                name=name,
                bindings=[HarlequinKeyBinding(**binding) for binding in bindings],
            )
        except TypeError as e:
            bad_key = str(e).split("argument ")[-1].strip("'")
            raise HarlequinConfigError(
                title="Harlequin could not load your keymap.",
                msg=(
                    "Key bindings must be defined in config files with "
                    "only three properties: `keys`, `action`, and `key_profile`. "
                    f"Got a binding in the map named {name} that tried to define "
                    f"a property: {bad_key!r}"
                ),
            ) from e
        return keymap

    def to_dict(self) -> RawKeyMap:
        return {"name": self.name, "bindings": [b.to_dict() for b in self.bindings]}



from __future__ import annotations

from typing import Union

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.validation import Integer
from textual.widget import Widget
from textual.widgets import Button, Checkbox, Input


class RunQueryBar(Horizontal):
    def __init__(
        self,
        *children: Widget,
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa
        classes: Union[str, None] = None,
        disabled: bool = False,
        max_results: int = 10_000,
        show_cancel_button: bool = False,
    ) -> None:
        self.max_results = max_results
        self.show_cancel_button = show_cancel_button
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        self.transaction_button = Button(
            "Tx: Auto", id="transaction_button", classes="hidden"
        )
        self.transaction_button.tooltip = "Click to change modes"
        self.commit_button = Button("🡅", id="commit_button", classes="hidden")
        self.commit_button.tooltip = "Commit transaction"
        self.rollback_button = Button("⮌", id="rollback_button", classes="hidden")
        self.rollback_button.tooltip = "Roll back transaction"
        self.limit_checkbox = Checkbox("Limit ", id="limit_checkbox")
        self.limit_input = Input(
            str(min(500, self.max_results)),
            id="limit_input",
            validators=Integer(
                minimum=0,
                maximum=self.max_results if self.max_results > 0 else None,
                failure_description=(
                    f"Please enter a number between 0 and {self.max_results}."
                    if self.max_results > 0
                    else "Please enter a number greater than 0."
                ),
            ),
        )
        self.run_button = Button("Run Query", id="run_query")
        self.cancel_button = Button("Cancel Query", id="cancel_query")
        self.cancel_button.add_class("hidden")
        with Horizontal(id="transaction_buttons"):
            yield self.transaction_button
            yield self.commit_button
            yield self.rollback_button
        with Horizontal(id="run_buttons"):
            yield self.limit_checkbox
            yield self.limit_input
            yield self.run_button
            yield self.cancel_button

    def on_mount(self) -> None:
        if self.app.is_headless:
            self.limit_input.cursor_blink = False

    @on(Input.Changed, "#limit_input")
    def handle_new_limit_value(self, message: Input.Changed) -> None:
        """
        check and uncheck the box for valid/invalid input
        """
        if (
            message.input.value
            and message.validation_result
            and message.validation_result.is_valid
        ):
            self.limit_checkbox.value = True
        else:
            self.limit_checkbox.value = False

    @property
    def limit_value(self) -> int | None:
        if not self.limit_checkbox.value or not self.limit_input.is_valid:
            return None
        try:
            return int(self.limit_input.value)
        except ValueError:
            return None

    def set_not_responsive(self) -> None:
        self.add_class("non-responsive")
        if self.show_cancel_button:
            with self.app.batch_update():
                self.run_button.add_class("hidden")
                self.cancel_button.remove_class("hidden")

    def set_responsive(self) -> None:
        self.remove_class("non-responsive")
        if self.show_cancel_button:
            with self.app.batch_update():
                self.run_button.remove_class("hidden")
                self.cancel_button.add_class("hidden")



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Sequence

from textual_fastdatatable.backend import AutoBackendType

from harlequin.autocomplete.completion import HarlequinCompletion
from harlequin.catalog import Catalog
from harlequin.options import HarlequinAdapterOption, HarlequinCopyFormat
from harlequin.transaction_mode import HarlequinTransactionMode


class HarlequinCursor(ABC):
    @abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def columns(self) -> list[tuple[str, str]]:
        """
        Gets a list of columns for the result set of the cursor.

        Returns: list[tuple[str, str]], where each tuple is the (name, type) of
            a column, where type should be a short (1-3 character) string. The
            columns must be ordered in the same order as the data returned by
            fetchall()
        """
        pass

    @abstractmethod
    def set_limit(self, limit: int) -> "HarlequinCursor":
        """
        Limits the number of results for future calls to fetchall().

        Args:
            limit (int): The maximum number of records to be returned
            by future calls to fetchall().

        Returns: HarlequinCursor, either a reference to self or a new
            cursor with the limit applied.
        """
        pass

    @abstractmethod
    def fetchall(self) -> AutoBackendType | None:
        """
        Returns data from the cursor's result set. Can return any type supported
        by textual-fastdatatable. If set_limit is called prior to fetchall,
        this method only returns the limited number of records. If the query returns
        no rows, fetchall should return None.

        Returns:
            pyarrow.Table |
            pyarrow.Record Batch |
            Sequence[Iterable[Any]] |
            Mapping[str, Sequence[Any]] |
            None
        """
        pass


class HarlequinConnection(ABC):
    """
    A Connection is created by the Adapter, and represents the main interface into
    the database.
    """

    @abstractmethod
    def __init__(self, *args: Any, init_message: str = "", **kwargs: Any) -> None:
        """
        Args:
            init_message (str): If set, Harlequin will notify the user with the
            message after the connection is created.
        """
        self.init_message = init_message

    @abstractmethod
    def execute(self, query: str) -> HarlequinCursor | None:
        """
        Executes query and returns a cursor (for a select stmt) or None. Raises
        HarlequinQueryError if the database raises an error in response to the query.

        Args:
            query (str): The text of a single query to execute

        Returns: HarlequinCursor | None

        Raises: HarlequinQueryError
        """
        pass

    def cancel(self) -> None:
        """
        Cancels/interrupts all in-flight queries previously executed with `execute`.

        After implementing this method, set the adapter class variable
        IMPLEMENTS_CANCEL to True to show the cancel button in the Harlequin UI.
        """
        return None

    @abstractmethod
    def get_catalog(self) -> Catalog:
        """
        Introspects the connected database and returns a Catalog object with
        items for each database, schema, table, view, column, etc.

        Returns: Catalog
        """
        pass

    def get_completions(self) -> list[HarlequinCompletion]:
        """
        Returns a list of extra completions to make available to the Query Editor's
        autocomplete. These could be dialect-specific functions, keywords, etc.
        Harlequin ships with a basic list of common ANSI-SQL keywords and functions.

        It is not necessary to provide completions for Catalog items, since Harlequin
        will create those from the Catalog.

        Returns: list[HarlequinCompletion]
        """
        return []

    def copy(
        self, query: str, path: Path, format_name: str, options: dict[str, Any]
    ) -> None:
        """
        Exports data returned by query to a file or directory at path, using
        options.
        Args:
            query (str): The text of the query (select stmt) to be executed.
            path (Path): The destination location for the file(s) to be written.
            format_name (str): The name of the selected export format.
            options (dict[str, Any]): A dict of format option names and values.

        Returns: None

        Raises:
            NotImplementedError if the adapter does not have copy functionality.
            HarlequinCopyError for all other exceptions during export.
        """
        raise NotImplementedError

    def validate_sql(self, text: str) -> str:
        """
        Parses text as one or more queries; returns text if parsing does not result
        in an error; otherwise returns the empty string ("").

        Args:
            text (str): The text, which may compose one or more queries and partial
                queries.

        Returns: str, either the original text or the empty string ("")

        Raises: NotImplementedError if the adapter does not provide this optional
            functionality.
        """
        raise NotImplementedError

    def close(self) -> None:
        """
        Closes the connection, if necessary. This function is called when the app
        quits.

        Returns: None
        """
        return None

    @property
    def transaction_mode(self) -> HarlequinTransactionMode | None:
        """
        The user-facing label of the currently-active transaction mode.

        Returns None if the adapter does not support different
        transaction modes.

        Returns: HarlequinTransactionMode | None
        """
        return None

    def toggle_transaction_mode(self) -> HarlequinTransactionMode | None:
        """
        Switches to the next transaction mode in the adapter's sequence of modes
        and returns the new mode.

        No-ops and returns None if the adapter does not support different
        transaction modes.

        Returns: HarlequinTransactionMode, the new mode, after toggling, or None.
        """
        return None


class HarlequinAdapter(ABC):
    """
    A HarlequinAdapter is the main abstraction for a database backend for
    Harlequin.

    It must declare its configuration setting the ADAPTER_OPTIONS
    class variable.

    Adapters are initialized with a conn_str (a tuple of strings), and
    kwargs that represent CLI options. Adapters must be robust to receiving
    both subsets and supersets of their defined options as kwargs. They should
    disregard any extra (unexpected) kwargs. They should not rely on the
    option's default values, as those will not be passed by the CLI when
    initializing the adapter.
    """

    ADAPTER_OPTIONS: list[HarlequinAdapterOption] | None = None
    COPY_FORMATS: list[HarlequinCopyFormat] | None = None
    """DEPRECATED. Adapter Copy formats are now ignored by Harlequin."""
    IMPLEMENTS_CANCEL = False

    @abstractmethod
    def __init__(self, conn_str: Sequence[str], **options: Any) -> None:
        """
        Initialize an adapter.

        Args:
            - conn_str (Sequence[str]): One or more connection strings. May be empty.
            - **options (Any): Options received from the command line, config file,
                or env variables. Adapters should be robust to receiving both subsets
                and supersets of their declared options. They should disregard any
                extra (unexpected) kwargs. Adapters should check the types of options,
                as they may not be cast to the correct types.

        Raises: HarlequinConfigError if a received option is the wrong value or type.
        """
        pass

    @abstractmethod
    def connect(self) -> HarlequinConnection:
        """
        Creates and returns an initialized connection to a database. Necessary config
        should be stored in the HarlequinAdapter instance when it is created.

        Returns: HarlequinConnection.

        Raises: HarlequinConnectionError if a connection could not be established.
        """
        pass

    @property
    def connection_id(self) -> str | None:
        """
        Returns a unique ID for this connection, typically a fully-hydrated connection
        string or similar. This Unique ID is used by Harlequin to cache the data
        catalog and query history and persist them across invocations of Harlequin.

        If None is returned, Harlequin will attempt to compute a unique ID from the
        arguments used to initialize the adapter.

        If the empty string is returned, Harlequin will not attempt to load the
        catalog or buffers from the cache.

        Returns: str | None
        """
        return None

    @property
    def implements_copy(self) -> bool:
        """
        True if the adapter's connection implements the copy() method. Adapter must
        also provide options for customizing the Export dialog GUI.
        """
        return self.COPY_FORMATS is not None



import tarfile
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from urllib.request import urlopen

import platformdirs
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.lib as pl

from harlequin.exception import HarlequinTzDataError, pretty_print_warning

HARLEQUIN_TZ_DATA_PATH = platformdirs.user_data_path(appname="harlequin") / "tzdata"


def check_and_install_tzdata() -> None:
    """
    On Windows, Arrow expects to find a timezone database in the User's
    Downloads folder. We check to see if it can find one there,
    and if not, we override the tz db search path to a Harlequin-specific
    location. If it still can't find one, we download the DB to that
    harlequin-specific location.
    """
    try:
        pc.assume_timezone(datetime(2024, 1, 1), "America/New_York")
    except pl.ArrowInvalid:
        # no tz database in the default location; try the harlequin location
        try:
            pa.set_timezone_db_path(str(HARLEQUIN_TZ_DATA_PATH))
            pc.assume_timezone(datetime(2024, 1, 1), "America/New_York")
        except (OSError, pl.ArrowInvalid):
            message = (
                "Harlequin could not find a timezone database, which it needs "
                "to support Apache Arrow's timestamptz features. It is downloading one "
                "now (it will only do this once). For more info, see "
                "[link]https://harlequin.sh/docs/troubleshooting/timezone-windows[/]"
            )
            pretty_print_warning(title="Harlequin Timezone Support", message=message)
            try:
                # tz database is missing. We need to install it.
                response = urlopen(
                    "https://www.iana.org/time-zones/repository/tzdata-latest.tar.gz"
                )
                with TemporaryDirectory() as tmpdir:
                    tar_path = Path(tmpdir) / "tzdata.tar.gz"
                    with tar_path.open("wb") as f:
                        f.write(response.read())
                    tarfile.open(tar_path).extractall(HARLEQUIN_TZ_DATA_PATH)
                zone_response = urlopen(
                    "https://raw.githubusercontent.com/unicode-org/cldr/main/common/"
                    "supplemental/windowsZones.xml"
                )
                zone_target = HARLEQUIN_TZ_DATA_PATH / "windowsZones.xml"
                with zone_target.open("wb") as f:
                    f.write(zone_response.read())
                pa.set_timezone_db_path(str(HARLEQUIN_TZ_DATA_PATH))
            except Exception as e:
                err_msg = (
                    "Harlequin was not able to download a timezone database. Without "
                    "a timezone database, Harlequin may crash if you attempt to load "
                    "timestamptz values into the results viewer. To use Harlequin "
                    "anyway, set the --no-download-tzdata option.\n"
                    "For more info, see "
                    "[link]https://harlequin.sh/docs/troubleshooting/timezone-windows[/]"
                    f"\nDownload failed with the following exception:\n{e}"
                )
                raise HarlequinTzDataError(
                    msg=err_msg, title="Harlequin Timezone Error"
                ) from e



from __future__ import annotations

from asyncio import PriorityQueue
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Generator, Iterable, TypeVar

from rich.style import Style
from rich.text import Text, TextType
from textual import work
from textual.await_complete import AwaitComplete
from textual.reactive import var
from textual.widgets._tree import Tree, TreeNode
from textual.worker import WorkerCancelled, WorkerFailed, get_current_worker

from harlequin.catalog import (
    Catalog,
    CatalogItem,
    InteractiveCatalogItem,
    NewCatalogItems,
)
from harlequin.components.data_catalog.tree import HarlequinTree
from harlequin.messages import WidgetMounted

if TYPE_CHECKING:
    from typing_extensions import Self


class DatabaseTree(HarlequinTree[CatalogItem], inherit_bindings=False):
    catalog: var[Catalog | None] = var["Catalog | None"](
        None, init=False, always_update=True
    )

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,  # noqa: A002
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self._load_queue: PriorityQueue[
            tuple[int, str, int, TreeNode[InteractiveCatalogItem]]
        ] = PriorityQueue()
        """
        _load_queue is a priority queue, ordered by a priority int,
        then the string label of the node, then the id of the node,
        which should prevent any comparisons between TreeNodes, which
        do not implement cmp operators.
        """
        super().__init__(
            label="Root",
            data=CatalogItem(
                qualified_identifier="__root__",
                query_name="",
                label="Root",
                type_label="",
                children=[],
            ),
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

    def on_mount(self) -> None:
        self.loading = True
        self.show_root = False
        self.guide_depth = 3
        self.root.expand()
        self.post_message(WidgetMounted(widget=self))

    def _build_item_label(self, label: str, type_label: str) -> Text:
        type_label_style = self.get_component_rich_style("harlequin-tree--type-label")
        type_label_fg_style = Style(color=type_label_style.color)
        return Text.assemble(label, " ", (type_label, type_label_fg_style))

    async def watch_catalog(self, catalog: Catalog | None) -> None:
        """Watch for changes to the `catalog` of the database tree.

        If the Catalog is changed the directory tree will be repopulated using
        the new value as the root.
        """
        assert isinstance(self.root.data, CatalogItem)
        self.root.data.children = catalog.items if catalog is not None else []
        await self.reload()
        # add the root's children to the prefetch queue
        for child in self.root.children:
            if isinstance(child.data, InteractiveCatalogItem) and not child.data.loaded:
                self._add_to_load_queue(child)  # type: ignore[arg-type]
        self.loading = False

    def _add_to_load_queue(
        self, node: TreeNode[InteractiveCatalogItem], priority: int = 100
    ) -> AwaitComplete:
        """Add the given node to the load priority queue.

        The return value can optionally be awaited until the queue is empty.

        Args:
            node: The node to add to the load queue.
            priority: An order for this node to be loaded; lowest first.

        Returns:
            An optionally awaitable object that can be awaited until the
                load queue has finished processing.
        """
        assert node.data is not None
        if not node.data.loaded:
            with suppress(TypeError):
                # typeError will be raised if this node already exists in the queue,
                # since TreeNodes do not implement cmp operators.
                self._load_queue.put_nowait((priority, str(node.label), id(node), node))

        return AwaitComplete(self._load_queue.join())

    def reload(self) -> AwaitComplete:
        """Reload the `DirectoryTree` contents.

        Returns:
            An optionally awaitable that ensures the tree has finished reloading.
        """
        # Orphan the old queue...
        self._load_queue = PriorityQueue()
        # ... reset the root node ...
        processed = self.reload_node(self.root)
        # ... and replace the old loader with a new one.
        self._loader()
        return processed

    def clear_node(self, node: TreeNode[CatalogItem]) -> Self:
        """Clear all nodes under the given node.

        Returns:
            The `Tree` instance.
        """
        self._clear_line_cache()
        node.remove_children()
        self._updates += 1
        self.refresh()
        return self

    def reset_node(
        self,
        node: TreeNode[CatalogItem],
        label: TextType,
        data: CatalogItem | None = None,
    ) -> Self:
        """Clear the subtree and reset the given node.

        Args:
            node: The node to reset.
            label: The label for the node.
            data: Optional data for the node.

        Returns:
            The `Tree` instance.
        """
        self.clear_node(node)
        node.label = label
        node.data = data
        return self

    async def _reload(self, node: TreeNode[CatalogItem]) -> None:
        """Reloads the subtree rooted at the given node while preserving state.

        After reloading the subtree, nodes that were expanded and still exist
        will remain expanded and the highlighted node will be preserved, if it
        still exists. If it doesn't, highlighting goes up to the first parent
        directory that still exists.

        Args:
            node: The root of the subtree to reload.
        """
        async with self.lock:
            # Track nodes that were expanded before reloading.
            currently_open: set[str] = set()
            to_check: list[TreeNode[CatalogItem]] = [node]
            while to_check:
                checking = to_check.pop()
                if checking.allow_expand and checking.is_expanded:
                    if checking.data:
                        currently_open.add(checking.data.qualified_identifier)
                    to_check.extend(checking.children)

            # Track node that was highlighted before reloading.
            highlighted_identifier: None | str = None
            if self.cursor_line > -1:
                highlighted_node = self.get_node_at_line(self.cursor_line)
                if highlighted_node is not None and highlighted_node.data is not None:
                    highlighted_identifier = highlighted_node.data.qualified_identifier

            if node.data is not None:
                self.reset_node(
                    node,
                    self._build_item_label(node.data.label, node.data.type_label),
                    node.data,
                )

            # Reopen nodes that were expanded and still exist.
            to_reopen = [node]
            while to_reopen:
                reopening = to_reopen.pop()
                if not reopening.data:
                    continue
                if reopening.allow_expand and (
                    reopening.data.qualified_identifier in currently_open
                    or reopening == node
                ):
                    try:
                        content = await self._load_children(reopening).wait()
                    except (WorkerCancelled, WorkerFailed):
                        continue
                    self._populate_node(reopening, content)
                    to_reopen.extend(reopening.children)
                    reopening.expand()

            if highlighted_identifier is None:
                return

            # Restore the highlighted path and consider the parents as fallbacks.
            looking = [node]

            def parents(qualified_identifier: str) -> Generator[str, None, None]:
                parent = qualified_identifier
                while parent := parent.rpartition(".")[0]:
                    yield parent
                yield "__root__"

            highlight_candidates = set(parents(highlighted_identifier))
            highlight_candidates.add(highlighted_identifier)
            best_found: None | TreeNode[CatalogItem] = None
            while looking:
                checking = looking.pop()
                checking_path = (
                    checking.data.qualified_identifier
                    if checking.data is not None
                    else None
                )
                if checking_path in highlight_candidates:
                    best_found = checking
                    if checking_path == highlighted_identifier:
                        break
                    if checking.allow_expand and checking.is_expanded:
                        looking.extend(checking.children)
            if best_found is not None:
                # We need valid lines. Make sure the tree lines have been computed:
                _ = self._tree_lines
                self.cursor_line = best_found.line

    def reload_node(self, node: TreeNode[CatalogItem]) -> AwaitComplete:
        """Reload the given node's contents.

        The return value may be awaited to ensure the DirectoryTree has reached
        a stable state and is no longer performing any node reloading (of this node
        or any other nodes).

        Args:
            node: The root of the subtree to reload.

        Returns:
            An optionally awaitable that ensures the subtree has finished reloading.
        """
        return AwaitComplete(self._reload(node))

    TCatalogItem_co = TypeVar("TCatalogItem_co", bound=CatalogItem, covariant=True)

    def _populate_node(
        self, node: TreeNode[TCatalogItem_co], content: Iterable[TCatalogItem_co]
    ) -> None:
        """Populate the given tree node with the given directory content.

        Args:
            node: The Tree node to populate.
            content: The collection of `Path` objects to populate the node with.
        """
        node.remove_children()
        for item in content:
            node.add(
                self._build_item_label(item.label, item.type_label),
                data=item,
                allow_expand=bool(item.children) or not getattr(item, "loaded", True),
            )

    @work(thread=True, exit_on_error=False, description="_load_children")
    def _load_children(self, node: TreeNode[CatalogItem]) -> list[CatalogItem]:
        """Load the children for a given node.

        Args:
            node: The node to load the children for.

        Returns:
            The list of entries within the directory associated with the node.
        """
        assert node.data is not None
        if (
            not node.data.children
            and isinstance(node.data, InteractiveCatalogItem)
            and not node.data.loaded
        ):
            try:
                children = list(node.data.fetch_children())
            except BaseException as e:
                self.post_message(self.CatalogError(catalog_type="database", error=e))
                return []
            else:
                node.data.children = children
                self.post_message(NewCatalogItems(parent=node.data, items=children))
            finally:
                node.data.loaded = True
                node.allow_expand = bool(node.data.children)

        return sorted(
            node.data.children,
            key=lambda catalog_item: catalog_item.label,
        )

    @work(name="_database_tree_background_loader", exclusive=True)
    async def _loader(self) -> None:
        """Background loading queue processor."""
        worker = get_current_worker()
        while not worker.is_cancelled:
            # Get the next node that needs loading off the queue. Note that
            # this blocks if the queue is empty.
            *_, node = await self._load_queue.get()
            content: list[Path] = []
            async with self.lock:
                try:
                    # Spin up a short-lived thread that will load the content of
                    # the directory associated with that node.
                    content = await self._load_children(node).wait()
                except WorkerCancelled:
                    # The worker was cancelled, that would suggest we're all
                    # done here and we should get out of the loader in general.
                    break
                except WorkerFailed:
                    # This particular worker failed to start. We don't know the
                    # reason so let's no-op that (for now anyway).
                    pass
                else:
                    # We're still here and we have directory content, get it into
                    # the tree.
                    if content:
                        self._populate_node(node, content)
                finally:
                    # Mark this iteration as done.
                    self._load_queue.task_done()

    async def _on_tree_node_expanded(
        self, event: Tree.NodeExpanded[CatalogItem]
    ) -> None:
        event.stop()
        node = event.node
        if node.data is None:
            return
        if isinstance(node.data, InteractiveCatalogItem) and not node.data.loaded:
            # if this node isn't loaded yet, add it to the front of the queue
            await self._add_to_load_queue(node, priority=0)  # type: ignore[arg-type]
        if (
            isinstance(node.data, CatalogItem)
            and node.data.children
            and not node.children
        ):
            self._populate_node(node, content=node.data.children)
        # pre-fetch the node's grandchildren
        for child in node.children:
            if isinstance(child.data, InteractiveCatalogItem):
                self._add_to_load_queue(child)  # type: ignore[arg-type]



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable


@dataclass
class HarlequinTransactionMode:
    """
    A container for a database connection transaction mode. A mode must
    have a label, and may also define a zero-arg callable that commits
    or rolls back a transaction (if these callables are defined, Harlequin
    will present buttons for committing and rolling back transactions).

    Args:
        label (str): A short label for this mode, to be shown in the UI
            as f"Tx: {label}"
        commit (Callable[[], None] | None): A callable to commit an open
            transaction.
        rollback (Callable[[], None] | None): A callable to roll back an
            open transaction.
    """

    label: str
    commit: Callable[[], None] | None = None
    rollback: Callable[[], None] | None = None



from __future__ import annotations

from harlequin.options import (
    FlagOption,
    HarlequinCopyFormat,
    SelectOption,
    TextOption,
)


def _validate_int(raw: str) -> tuple[bool, str | None]:
    try:
        int(raw)
    except ValueError:
        return False, "Must be an integer."
    else:
        return True, None


def _validate_int_or_empty(raw: str) -> tuple[bool, str | None]:
    if raw == "":
        return True, None
    else:
        return _validate_int(raw)


def _validate_float(raw: str) -> tuple[bool, str | None]:
    try:
        float(raw)
    except ValueError:
        return False, "Must be a floating point number."
    else:
        return True, None


csv = HarlequinCopyFormat(
    name="csv",
    label="CSV",
    extensions=(".csv", ".tsv"),
    options=[
        FlagOption(
            name="header",
            description="Switch on to include column name headers.",
            label="Header",
        ),
        TextOption(
            name="sep",
            description="The separator (or delimeter) between cols in each row.",
            label="Separator",
            default=",",
        ),
        SelectOption(
            name="compression",
            description=(
                "The compression type for the file. By default this will be detected "
                "automatically from the file extension (e.g. file.csv.gz will use "
                "gzip, file.csv will use no compression)."
            ),
            default="auto",
            choices=[
                ("Auto", "auto"),
                ("gzip", "gzip"),
                ("zstd", "zstd"),
                ("No compression", "none"),
            ],
        ),
        FlagOption(
            name="quoting",
            description="Switch on to always quote all strings.",
            label="Force Quote",
        ),
        TextOption(
            name="date_format",
            description="Specifies the date format to use when writing dates.",
            label="Date Format",
            default="",
            placeholder="%Y-%m-%d",
        ),
        TextOption(
            name="timestamp_format",
            description="Specifies the date format to use when writing timestamps.",
            label="Timestamp Format",
            default="",
            placeholder="%c",
        ),
        TextOption(
            name="quotechar",
            description="The quoting character to be used when a data value is quoted.",
            label="Quote Char",
            default='"',
        ),
        TextOption(
            name="escapechar",
            description=(
                "The character that should appear before a character that matches the "
                "quote value."
            ),
            label="Escape Char",
            default='"',
        ),
        TextOption(
            name="na_rep",
            description="The string that is written to represent a NULL value.",
            label="Null String",
            default="",
        ),
        TextOption(
            name="encoding",
            description="Only UTF8 is currently supported by DuckDB.",
            label="Encoding",
            default="UTF8",
        ),
    ],
)


parquet = HarlequinCopyFormat(
    name="parquet",
    label="Parquet",
    extensions=(".parquet", ".pq"),
    options=[
        SelectOption(
            name="compression",
            description=(
                "The compression format to use (uncompressed, snappy, gzip or zstd). "
                "Default snappy."
            ),
            choices=[
                ("Snappy", "snappy"),
                ("gzip", "gzip"),
                ("zstd", "zstd"),
                ("Uncompressed", "uncompressed"),
            ],
            default="snappy",
        )
    ],
)


json = HarlequinCopyFormat(
    name="json",
    label="JSON",
    extensions=(".json", ".js", ".ndjson"),
    options=[
        FlagOption(
            name="array",
            description=(
                "Whether to write a JSON array. If true, a JSON array of records is "
                "written, if false, newline-delimited JSON is written."
            ),
        ),
        SelectOption(
            name="compression",
            description=(
                "The compression type for the file. By default this will be detected "
                "automatically from the file extension (e.g. file.json.gz will use "
                "gzip, file.json will use no compression)."
            ),
            choices=[
                ("Auto", "auto"),
                ("gzip", "gzip"),
                ("zstd", "zstd"),
                ("Uncompressed", "uncompressed"),
            ],
            default="auto",
        ),
        TextOption(
            name="date_format",
            description="Specifies the date format to use when writing dates.",
            label="Date Format",
            default="",
            placeholder="%Y-%m-%d",
        ),
        TextOption(
            name="timestamp_format",
            description="Specifies the date format to use when writing timestamps.",
            label="Timestamp Format",
            default="",
            placeholder="%c",
        ),
    ],
)

orc = HarlequinCopyFormat(
    name="orc",
    label="ORC",
    extensions=(".orc",),
    options=[
        SelectOption(
            name="file_version",
            description=(
                "Determine which ORC file version to use. Hive 0.11 / ORC v0 is the "
                "older version while Hive 0.12 / ORC v1 is the newer one."
            ),
            label="File Version",
            choices=[
                ("0.11", "0.11"),
                ("0.12", "0.12"),
            ],
            default="0.12",
        ),
        TextOption(
            name="batch_size",
            description="Number of rows the ORC writer writes at a time.",
            label="Batch Size",
            default="1024",
            validator=_validate_int,
        ),
        TextOption(
            name="stripe_size",
            description="Size of each ORC stripe in bytes.",
            label="Stripe Size",
            default=str(64 * 1024 * 1024),
            validator=_validate_int,
        ),
        SelectOption(
            name="compression",
            label="Compression",
            description="The compression codec.",
            choices=[
                ("Uncompressed", "UNCOMPRESSED"),
                ("Snappy", "SNAPPY"),
                ("zlib", "ZLIB"),
                ("LZ4", "LZ4"),
                ("zstd", "zstd"),
            ],
            default="UNCOMPRESSED",
        ),
        TextOption(
            name="compression_block_size",
            label="Compression Block Size",
            description="Size of each compression block in bytes.",
            default=str(64 * 1024),
            validator=_validate_int,
        ),
        SelectOption(
            name="compression_strategy",
            label="Compression Strategy",
            description="The compression strategy i.e. speed vs size reduction.",
            choices=[
                ("Speed", "SPEED"),
                ("Compression", "COMPRESSION"),
            ],
            default="SPEED",
        ),
        TextOption(
            name="row_index_stride",
            description=(
                "The row index stride i.e. the number of rows per an entry in the "
                "row index."
            ),
            label="Row Index Stride",
            default="10000",
            validator=_validate_int,
        ),
        TextOption(
            name="padding_tolerance",
            label="Padding Tolerance",
            description="The padding tolerance.",
            default="0.0",
            validator=_validate_float,
        ),
        TextOption(
            name="dictionary_key_size_threshold",
            label="Dict Key Size Threshold",
            description=(
                "The dictionary key size threshold. 0 to disable dictionary encoding. "
                "1 to always enable dictionary encoding."
            ),
            default="0.0",
            validator=_validate_float,
        ),
        TextOption(
            name="bloom_filter_columns",
            label="Bloom Filter Columns",
            description="Columns that use the bloom filter. Separate with a comma",
        ),
        TextOption(
            name="bloom_filter_fpp",
            label="Bloom Filter False-Positive",
            description=("Upper limit of the false-positive rate of the bloom filter."),
            default="0.05",
            validator=_validate_float,
        ),
    ],
)

feather = HarlequinCopyFormat(
    name="feather",
    label="Feather",
    extensions=(".feather",),
    options=[
        SelectOption(
            name="compression",
            label="Compression",
            description="The compression codec.",
            choices=[
                ("Uncompressed", "uncompressed"),
                ("LZ4", "lz4"),
                ("zstd", "zstd"),
            ],
            default="uncompressed",
        ),
        TextOption(
            name="compression_level",
            label="Compression Level",
            description=(
                "Use a compression level particular to the chosen compressor. "
                "If None use the default compression level."
            ),
            default="",
            validator=_validate_int_or_empty,
        ),
        TextOption(
            name="chunksize",
            label="Chunk Size",
            description=(
                "For V2 files, the internal maximum size of Arrow RecordBatch chunks "
                "when writing the Arrow IPC file format. None means use the default, "
                "which is currently 64K."
            ),
            default="",
            validator=_validate_int_or_empty,
        ),
        SelectOption(
            name="version",
            description=(
                "Feather file version. Version 2 is the current. Version 1 is the "
                "more limited legacy format."
            ),
            label="File Version",
            choices=[
                ("2", "2"),
                ("1", "1"),
            ],
            default="2",
        ),
    ],
)

HARLEQUIN_COPY_FORMATS = [csv, parquet, json, orc, feather]
WINDOWS_COPY_FORMATS = [csv, parquet, json, feather]



from __future__ import annotations

from typing import ClassVar
from urllib.parse import urlsplit

from textual import on, work
from textual.message import Message
from textual.widgets import (
    DirectoryTree,
)
from textual.widgets._tree import TreeNode
from textual.worker import Worker, WorkerState

from harlequin.catalog_cache import recursive_dict
from harlequin.components.data_catalog.tree import HarlequinTree as HarlequinTree
from harlequin.messages import WidgetMounted

try:
    import boto3
except ImportError:
    boto3 = None  # type: ignore


class S3Tree(HarlequinTree[str], inherit_bindings=False):
    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "directory-tree--extension",
        "directory-tree--file",
        "directory-tree--folder",
        "directory-tree--hidden",
    }

    class DataReady(Message):
        def __init__(self, data: dict) -> None:
            self.data = data
            super().__init__()

    def __init__(
        self,
        uri: str,
        data: str | None = None,
        name: str | None = None,
        id: str | None = None,  # noqa: A002
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.endpoint_url, self.bucket, self.prefix = self._parse_s3_uri(uri)
        self.catalog_data: dict | None = None
        super().__init__(
            "Root", data, name=name, id=id, classes=classes, disabled=disabled
        )

    ICON_NODE_EXPANDED = DirectoryTree.ICON_NODE_EXPANDED
    ICON_NODE = DirectoryTree.ICON_NODE
    ICON_FILE = DirectoryTree.ICON_FILE

    render_label = DirectoryTree.render_label  # type: ignore[assignment]

    def on_mount(self) -> None:
        self.guide_depth = 3
        self.show_root = False
        self.root.data = self.endpoint_url or "s3:/"
        self.reload()
        self.post_message(WidgetMounted(widget=self))

    @property
    def cache_key(self) -> tuple[str | None, str | None, str | None]:
        return (self.endpoint_url, self.bucket, self.prefix)

    @on(DataReady)
    def build_tree_from_message_data(self, message: S3Tree.DataReady) -> None:
        self.build_tree(message.data)

    def build_tree(self, data: dict) -> None:
        self.catalog_data = data
        self.clear()
        self._build_subtree(data=data, parent=self.root)
        self.root.expand()
        self.loading = False

    @on(Worker.StateChanged)
    async def handle_worker_state_change(self, message: Worker.StateChanged) -> None:
        if message.state == WorkerState.ERROR:
            await self._handle_worker_error(message)

    async def _handle_worker_error(self, message: Worker.StateChanged) -> None:
        if (
            message.worker.name == "_reload_objects"
            and message.worker.error is not None
        ):
            self.post_message(
                self.CatalogError(catalog_type="s3", error=message.worker.error)
            )
            self.loading = False

    @staticmethod
    def _parse_s3_uri(uri: str) -> tuple[str | None, str | None, str | None]:
        """
        Any of these are acceptable:
        my-bucket
        my-bucket/my-prefix
        s3://my-bucket
        s3://my-bucket/my-prefix
        https://my-storage.com/my-bucket/
        https://my-storage.com/my-bucket/my-prefix
        https://my-bucket.s3.amazonaws.com/my-prefix
        https://my-bucket.storage.googleapis.com/my-prefix
        """

        def _is_prefixed_aws_url(netloc: str) -> bool:
            parts = netloc.split(".")
            print(parts)
            if ".".join(parts[1:]) == "s3.amazonaws.com":
                return True
            return False

        def _is_prefixed_gcs_url(netloc: str) -> bool:
            parts = netloc.split(".")
            if ".".join(parts[1:]) == "storage.googleapis.com":
                return True
            return False

        if uri == "all":
            # special keyword so we list all buckets
            return None, None, None

        scheme, netloc, path, *_ = urlsplit(uri)
        path = path.lstrip("/")
        bucket: str | None

        if not scheme:
            assert not netloc
            endpoint_url = None
            bucket = path.split("/")[0]
            prefix = "/".join(path.split("/")[1:])
        elif scheme == "s3":
            endpoint_url = None
            bucket = netloc
            prefix = path
        elif _is_prefixed_aws_url(netloc):
            endpoint_url = None
            bucket = netloc.split(".")[0]
            prefix = path
        elif _is_prefixed_gcs_url(netloc):
            endpoint_url = "https://storage.googleapis.com"
            bucket = netloc.split(".")[0]
            prefix = path
        else:
            endpoint_url = f"{scheme}://{netloc}"
            bucket = path.split("/")[0] or None
            prefix = "/".join(path.split("/")[1:])

        return endpoint_url, bucket, prefix

    def reload(self) -> None:
        self.loading = True
        self._reload_objects()

    @work(thread=True, exclusive=True, exit_on_error=False)
    def _reload_objects(self) -> None:
        if boto3 is None:
            return

        data = {}
        s3 = boto3.resource("s3", endpoint_url=self.endpoint_url)
        if self.bucket is None:
            buckets = [b for b in s3.buckets.all()]
        else:
            buckets = [s3.Bucket(self.bucket)]
        for bucket in buckets:
            data[bucket.name] = recursive_dict()
            object_gen = (
                bucket.objects.filter(Prefix=self.prefix)
                if self.prefix
                else bucket.objects.all()
            )
            for obj in object_gen:
                key_parts = obj.key.split("/")
                target = data[bucket.name]
                for part in key_parts:
                    target = target[part]

        self.post_message(S3Tree.DataReady(data=data))

    def _build_subtree(
        self,
        data: dict[str, dict],
        parent: TreeNode[str],
    ) -> None:
        for k in data:
            if data[k]:
                new_node = parent.add(
                    label=k,
                    data=f"{parent.data}/{k}",
                )
                self._build_subtree(data[k], new_node)
            else:
                parent.add_leaf(label=k, data=f"{parent.data}/{k}")



from __future__ import annotations

from pathlib import Path
from typing import Any

import questionary
import tomlkit
from rich import print as rich_print
from rich.markup import escape
from rich.panel import Panel
from textual.theme import BUILTIN_THEMES

from harlequin.adapter import HarlequinAdapter
from harlequin.colors import HARLEQUIN_QUESTIONARY_STYLE, YELLOW
from harlequin.config import (
    Config,
    ConfigFile,
    Profile,
    get_config_for_profile,
    get_highest_priority_existing_config_file,
    sluggify_option_name,
)
from harlequin.exception import HarlequinWizardError, pretty_print_error
from harlequin.options import ListOption
from harlequin.plugins import load_adapter_plugins, load_keymap_plugins


def wizard(config_path: Path | None) -> None:
    try:
        _wizard(config_path)
    except KeyboardInterrupt:
        print("Cancelled config updates. No changes were made to any files.")
        return
    except HarlequinWizardError as e:
        pretty_print_error(e)
        return


def _wizard(config_path: Path | None) -> None:
    path = _prompt_for_path(config_path)
    config_file = ConfigFile(path)
    config = config_file.relevant_config

    # extract existing profiles from config file.
    if "profiles" not in config:
        config["profiles"] = {}
    profiles = config["profiles"]

    profile_name = _prompt_for_profile_name(profiles)
    selected_profile = profiles.get(profile_name, {})

    adapters = load_adapter_plugins()
    adapter = questionary.select(
        message="Which adapter should this profile use?",
        choices=sorted(adapters.keys()),
        default=selected_profile.get("adapter", "duckdb"),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    conn_str = questionary.text(
        message="What connection string(s) should this profile use?",
        instruction="Separate items by a space.",
        default=" ".join(selected_profile.get("conn_str", [])),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    theme = questionary.select(
        message="What theme should this profile use?",
        choices=sorted(BUILTIN_THEMES.keys()),
        default=selected_profile.get("theme", "harlequin"),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    keymap_choices = [
        questionary.Choice(
            title=opt,
            checked=opt in selected_profile.get("keymap_name", ["vscode"]),
        )
        for opt in _all_keymap_names(config_path=config_path)
    ]

    keymap_name = questionary.checkbox(
        message="Which keymaps would you like to use?",
        choices=keymap_choices,
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    limit = int(
        questionary.text(
            message="How many rows should the data table show?",
            validate=_validate_int,
            default=str(selected_profile.get("limit", 100000)),
            style=HARLEQUIN_QUESTIONARY_STYLE,
        ).unsafe_ask()
    )

    show_files = questionary.path(
        message="Show local files from a directory? (Leave blank to hide)",
        validate=_validate_dir_or_blank,
        only_directories=True,
        default=str(selected_profile.get("show_files", "")),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    show_s3 = questionary.text(
        message="Show cloud storage files?",
        instruction=(
            "Enter bucket name or URI (or `all`), or leave blank to hide "
            "cloud storage viewer."
        ),
        default=str(selected_profile.get("show_s3", "")),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    locale = questionary.text(
        message="What locale should Harlequin use for formatting numbers?",
        instruction="Leave blank to use the system locale.",
        default=selected_profile.get("locale", ""),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    adapter_cls = adapters[adapter]
    adapter_option_choices = (
        [
            questionary.Choice(
                title=opt.name,
                checked=sluggify_option_name(opt.name) in selected_profile,
            )
            for opt in adapter_cls.ADAPTER_OPTIONS
        ]
        if adapter_cls.ADAPTER_OPTIONS is not None
        else []
    )

    if adapter_option_choices:
        which = questionary.checkbox(
            message="Which of the following adapter options would you like to set?",
            choices=adapter_option_choices,
            style=HARLEQUIN_QUESTIONARY_STYLE,
        ).unsafe_ask()
    else:
        which = []

    adapter_options = {}
    if conn_str:
        adapter_options["conn_str"] = conn_str.split(" ")
    _prompt_to_set_adapter_options(
        adapter_options=adapter_options,
        adapter_cls=adapter_cls,
        which=which,
        selected_profile=selected_profile,
    )

    default_profile = _prompt_to_set_default_profile(profile_name, config, profiles)

    new_profile: Profile = {
        "adapter": adapter,
        "theme": theme,
        "limit": limit,
        "keymap_name": keymap_name,
    }

    if show_files:
        new_profile["show_files"] = show_files

    if show_s3:
        new_profile["show_s3"] = show_s3

    if locale:
        new_profile["locale"] = locale

    new_profile.update(adapter_options)  # type: ignore[typeddict-item]

    _confirm_profile_generation(default_profile, profile_name, new_profile)

    config["profiles"][profile_name] = new_profile

    config_file.update(config=config)
    config_file.write()


def _prompt_for_path(config_path: Path | None) -> Path:
    if config_path is None:
        existing = get_highest_priority_existing_config_file()
        raw_path: str = questionary.path(
            "What config file do you want to create or update?",
            default=str(existing) if existing is not None else ".harlequin.toml",
            validate=lambda p: (
                True if p.endswith(".toml") else "Must have a .toml extension"
            ),
            style=HARLEQUIN_QUESTIONARY_STYLE,
        ).unsafe_ask()
        path = Path(raw_path).expanduser().resolve()
    else:
        path = config_path
        rich_print(
            f"[italic]Updating the file at [bold {YELLOW}]{escape(str(path))}"
            f"[/ bold {YELLOW}]:[/]"
        )
    if path.suffix != ".toml":
        raise HarlequinWizardError(
            msg="Must create a file with a .toml extension.",
            title="Harlequin could not create your configuration.",
        )
    return path


def _prompt_for_profile_name(profiles: dict[str, Profile]) -> str:
    NEW_PROFILE_SENTINEL = "[Create a New Profile]"
    profile_name = NEW_PROFILE_SENTINEL
    if profiles:
        profile_name = questionary.select(
            message="Which profile would you like to update?",
            choices=[NEW_PROFILE_SENTINEL, *profiles.keys()],
            style=HARLEQUIN_QUESTIONARY_STYLE,
        ).unsafe_ask()
    if profile_name == NEW_PROFILE_SENTINEL:
        profile_name = questionary.text(
            message="What would you like to name your profile?",
            style=HARLEQUIN_QUESTIONARY_STYLE,
            validate=lambda x: True if x and x != "None" else "Cannot be empty or None",
        ).unsafe_ask()
    return profile_name


def _prompt_to_set_adapter_options(
    adapter_options: dict[str, Any],
    adapter_cls: type[HarlequinAdapter],
    which: list[str],
    selected_profile: Profile,
) -> None:
    """
    Mutates passed adapter_options dict.
    """
    if which and adapter_cls.ADAPTER_OPTIONS is not None:
        for option in adapter_cls.ADAPTER_OPTIONS:
            if option.name not in which:
                continue
            value = option.to_questionary(
                selected_profile.get(sluggify_option_name(option.name), None)
            ).unsafe_ask()
            if isinstance(option, ListOption):
                value = value.split(" ")
            adapter_options.update({sluggify_option_name(option.name): value})


def _prompt_to_set_default_profile(
    profile_name: str, config: Config, profiles: dict[str, Profile]
) -> str | None:
    possible_names = set([profile_name, *profiles.keys()])
    NO_DEFAULT_SENTINEL = "[No default]"
    default_profile: str = questionary.select(
        message="Would you like to set a default profile?",
        choices=[
            NO_DEFAULT_SENTINEL,
            *possible_names,
        ],
        default=config.get("default_profile", None),
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).unsafe_ask()

    if default_profile == NO_DEFAULT_SENTINEL:
        _ = config.pop("default_profile", None)
        return None
    else:
        config["default_profile"] = default_profile
        return default_profile


def _confirm_profile_generation(
    default_profile: str | None, profile_name: str, new_profile: Profile
) -> None:
    new_config: Config = (
        {} if default_profile is None else {"default_profile": default_profile}
    )
    new_config.update({"profiles": {profile_name: new_profile}})
    new_config_toml = tomlkit.dumps(new_config).rstrip()

    rich_print("[italic] We generated the following profile:[/]")
    rich_print(
        Panel.fit(
            escape(new_config_toml),
            border_style=YELLOW,
        )
    )

    all_good = questionary.confirm(
        "Save this profile?",
        style=HARLEQUIN_QUESTIONARY_STYLE,
    ).ask()

    if not all_good:
        raise KeyboardInterrupt()


def _all_keymap_names(config_path: Path | None) -> list[str]:
    _, user_defined_keymaps = get_config_for_profile(
        config_path=config_path, profile_name=None
    )
    all_keymaps = load_keymap_plugins(user_defined_keymaps=user_defined_keymaps)
    return [k for k in all_keymaps.keys()]


def _validate_int(raw: str) -> bool:
    try:
        int(raw)
    except ValueError:
        return False
    else:
        return True


def _validate_dir_or_blank(raw: str) -> bool:
    if not raw:
        return True
    p = Path(raw)
    return p.exists() and p.is_dir()



from __future__ import annotations

import duckdb


def get_completion_data(
    cur: duckdb.DuckDBPyConnection,
) -> list[tuple[str, str, int, str | None]]:
    keyword_data = cur.execute(
        """
        select distinct
            keyword_name as label,
            'kw' as type_label,
            case when keyword_category = 'reserved' then 100 else 1000 end as priority,
            null as context
        from duckdb_keywords()
        """
    ).fetchall()

    function_data = cur.execute(
        """
        select distinct
            function_name as label,
            case when function_type = 'pragma' then 'pragma'
                when function_type = 'macro' then 'macro'
                when function_type = 'aggregate' then 'agg'
                when function_type = 'scalar' then 'fn'
                when function_type = 'table' then 'fn->T'
                else 'fn' end as type_label,
            1000 as priority,
            case 
                when database_name == 'system' then null 
                else schema_name 
            end as context
        from duckdb_functions()
        where database_name != 'temp'
        """
    ).fetchall()

    settings_data = cur.execute(
        """
        select distinct
            name as label,
            'set' as type_label,
            2000 as priority,
            null as context
        from duckdb_settings()
        """
    ).fetchall()

    type_data = cur.execute(
        """
        with
            system_types as (
                select distinct
                    type_name as label, 
                    'type' as type_label, 
                    1000 as priority, 
                    null as context
                from duckdb_types()
                where database_name = 'system'
            ),
            custom_types as (
                select distinct
                    type_name as label,
                    'type' as type_label,
                    1000 as priority,
                    schema_name as context
                from duckdb_types()
                where
                    database_name not in ('system', 'temp')
                    and type_name not in (select label from system_types)
            )
        select *
        from system_types
        union all
        select *
        from custom_types
        """
    ).fetchall()

    return [
        *keyword_data,
        *function_data,
        *settings_data,
        *type_data,
    ]



from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from datetime import datetime
from typing import Iterator

from rich.columns import Columns
from rich.console import Group, RenderableType
from rich.text import Text


@dataclass
class QueryExecution:
    query_text: str
    executed_at: datetime
    result_row_count: int
    elapsed: float

    def __rich__(self) -> RenderableType:
        ts = self.executed_at.strftime("%a, %b %d %H:%M:%S")
        if self.result_row_count < 0:
            result = Text("ERROR", style="bold italic red", justify="right")
        else:
            res = (
                f"{self.result_row_count:n} "
                f"{'record' if self.result_row_count == 1 else 'records'}"
                if self.result_row_count
                else "SUCCESS"
            )
            elapsed = f"{self.elapsed:.2f}s"
            result = Text.assemble(
                (res, "bold"), " in ", (elapsed, "bold"), justify="right"
            )
        query_lines = self.query_text.strip().splitlines()
        if len(query_lines) > 8:
            continuation: RenderableType = Text(
                f"… ({len(query_lines) - 7} more lines)\n", style="italic"
            )
            query_lines = query_lines[0:7]
        else:
            continuation = ""

        return Group(
            Columns(
                renderables=[Text(ts, style="bold"), result],
                expand=True,
            ),
            "\n".join(query_lines),
            continuation,
        )


@dataclass
class History:
    queries: deque[QueryExecution]

    def __iter__(self) -> Iterator[QueryExecution]:
        return iter(self.queries)

    def append(self, query_text: str, result_row_count: int, elapsed: float) -> None:
        self.queries.append(
            QueryExecution(
                query_text=query_text.strip(),
                executed_at=datetime.now(),
                result_row_count=result_row_count,
                elapsed=elapsed,
            )
        )

    @classmethod
    def blank(cls) -> "History":
        return cls(queries=deque([], maxlen=500))



from typing import TYPE_CHECKING

from textual.message import Message

if TYPE_CHECKING:
    from textual.widget import Widget


class WidgetMounted(Message):
    def __init__(self, widget: "Widget") -> None:
        super().__init__()
        self.widget = widget



from rich.panel import Panel


class HarlequinExit(Exception):
    pass


class HarlequinError(Exception):
    def __init__(self, msg: str, title: str = "") -> None:
        super().__init__(msg)
        self.msg = msg
        self.title = title


class HarlequinBindingError(HarlequinError):
    pass


class HarlequinConnectionError(HarlequinError):
    pass


class HarlequinCopyError(HarlequinError):
    pass


class HarlequinQueryError(HarlequinError):
    pass


class HarlequinThemeError(HarlequinError):
    pass


class HarlequinConfigError(HarlequinError):
    pass


class HarlequinWizardError(HarlequinError):
    pass


class HarlequinTzDataError(HarlequinError):
    pass


class HarlequinLocaleError(HarlequinError):
    pass


def pretty_print_error(error: HarlequinError) -> None:
    from rich import print as rich_print

    rich_print(pretty_error_message(error))


def pretty_error_message(error: HarlequinError) -> Panel:
    return Panel.fit(
        str(error),
        title=error.title if error.title else ("Harlequin encountered an error."),
        title_align="left",
        border_style="red",
    )


def pretty_print_warning(title: str, message: str) -> None:
    from rich import print as rich_print

    from harlequin.colors import GREEN

    rich_print(
        Panel.fit(
            message,
            title=title,
            title_align="left",
            border_style=GREEN,
        )
    )



from __future__ import annotations

import sqlite3

from harlequin.autocomplete.completion import HarlequinCompletion


def get_completion_data(conn: sqlite3.Connection) -> list[HarlequinCompletion]:
    KEYWORDS = [
        "abort",
        "action",
        "add",
        "after",
        "all",
        "alter",
        "always",
        "analyze",
        "and",
        "as",
        "asc",
        "attach",
        "autoincrement",
        "before",
        "begin",
        "between",
        "by",
        "cascade",
        "case",
        "cast",
        "check",
        "collate",
        "column",
        "commit",
        "conflict",
        "constraint",
        "create",
        "cross",
        "current",
        "current_date",
        "current_time",
        "current_timestamp",
        "database",
        "default",
        "deferrable",
        "deferred",
        "delete",
        "desc",
        "detach",
        "distinct",
        "do",
        "drop",
        "each",
        "else",
        "end",
        "escape",
        "except",
        "exclude",
        "exclusive",
        "exists",
        "explain",
        "fail",
        "filter",
        "first",
        "following",
        "for",
        "foreign",
        "from",
        "full",
        "generated",
        "glob",
        "group",
        "groups",
        "having",
        "if",
        "ignore",
        "immediate",
        "in",
        "index",
        "indexed",
        "initially",
        "inner",
        "insert",
        "instead",
        "intersect",
        "into",
        "is",
        "isnull",
        "join",
        "key",
        "last",
        "left",
        "like",
        "limit",
        "match",
        "materialized",
        "natural",
        "no",
        "not",
        "nothing",
        "notnull",
        "null",
        "nulls",
        "of",
        "offset",
        "on",
        "or",
        "order",
        "others",
        "outer",
        "over",
        "partition",
        "plan",
        "pragma",
        "preceding",
        "primary",
        "query",
        "raise",
        "range",
        "recursive",
        "references",
        "regexp",
        "reindex",
        "release",
        "rename",
        "replace",
        "restrict",
        "returning",
        "right",
        "rollback",
        "row",
        "rows",
        "savepoint",
        "select",
        "set",
        "table",
        "temp",
        "temporary",
        "then",
        "ties",
        "to",
        "transaction",
        "trigger",
        "unbounded",
        "union",
        "unique",
        "update",
        "using",
        "vacuum",
        "values",
        "view",
        "virtual",
        "when",
        "where",
        "window",
        "with",
        "without",
    ]

    PRAGMAS = [
        "analysis_limit",
        "application_id",
        "auto_vacuum",
        "automatic_index",
        "busy_timeout",
        "cache_size",
        "cache_spill",
        "cell_size_check",
        "checkpoint_fullfsync",
        "collation_list",
        "compile_options",
        "data_version",
        "database_list",
        "defer_foreign_keys",
        "encoding",
        "foreign_key_check",
        "foreign_key_list",
        "foreign_keys",
        "freelist_count",
        "fullfsync",
        "function_list",
        "hard_heap_limit",
        "ignore_check_constraints",
        "incremental_vacuum",
        "index_info",
        "index_list",
        "index_xinfo",
        "integrity_check",
        "journal_mode",
        "journal_size_limit",
        "legacy_alter_table",
        "legacy_file_format",
        "locking_mode",
        "max_page_count",
        "mmap_size",
        "module_list",
        "optimize",
        "page_count",
        "page_size",
        "pragma_list",
        "query_only",
        "quick_check",
        "read_uncommitted",
        "recursive_triggers",
        "reverse_unordered_selects",
        "secure_delete",
        "shrink_memory",
        "soft_heap_limit",
        "synchronous",
        "table_info",
        "table_list",
        "table_xinfo",
        "temp_store",
        "threads",
        "trusted_schema",
        "user_version",
        "wal_autocheckpoint",
        "wal_checkpoint",
    ]

    function_data = conn.execute(
        """
        select distinct
            name as label,
            case when type = 'w' then 'agg' else 'fn' end as type_label
        from pragma_function_list
        order by name asc
        """
    ).fetchall()

    keyword_completions = [
        HarlequinCompletion(
            label=kw_name, type_label="kw", value=kw_name, priority=100, context=None
        )
        for kw_name in KEYWORDS
    ]

    pragma_completions = [
        HarlequinCompletion(
            label=label, type_label="pragma", value=label, priority=1000, context=None
        )
        for label in PRAGMAS
    ]

    function_completions = [
        HarlequinCompletion(
            label=label, type_label=type_label, value=label, priority=1000, context=None
        )
        for label, type_label in function_data
    ]

    return [*keyword_completions, *pragma_completions, *function_completions]



from __future__ import annotations

from pathlib import Path
from typing import Any, Protocol

from harlequin.components.results_viewer import ResultsTable
from harlequin.exception import HarlequinCopyError


class ExporterCallable(Protocol):
    def __call__(self, table: ResultsTable, dest_path: str, **kwargs: Any) -> None: ...


def copy(
    table: ResultsTable, path: Path, format_name: str, options: dict[str, Any]
) -> None:
    assert table.backend is not None
    if table.row_count == 0:
        raise HarlequinCopyError("Cannot export empty table.")
    dest_path = str(path.expanduser())
    kwargs = {k: v for k, v in options.items() if v}
    exporters: dict[str, ExporterCallable] = {
        "csv": _export_csv,
        "parquet": _export_parquet,
        "json": _export_json,
        "orc": _export_orc,
        "feather": _export_feather,
    }
    exporters[format_name](table, dest_path, **kwargs)


def _export_csv(
    table: ResultsTable,
    dest_path: str,
    **kwargs: Any,
) -> None:
    import duckdb

    assert table and table.backend
    if kwargs.get("quoting"):
        kwargs["quoting"] = "ALL"
    try:
        relation = duckdb.arrow(table.backend.source_data)
        relation.write_csv(file_name=dest_path, **kwargs)
    except (duckdb.Error, OSError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("DuckDB raised an error when writing your query to a CSV file."),
        ) from e


def _export_parquet(
    table: ResultsTable,
    dest_path: str,
    **kwargs: Any,
) -> None:
    import duckdb

    assert table and table.backend

    try:
        relation = duckdb.arrow(table.backend.source_data)
        relation.write_parquet(
            file_name=dest_path, compression=kwargs.get("compression")
        )
    except (duckdb.Error, OSError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("DuckDB raised an error when writing your query to a Parquet file."),
        ) from e


def _export_json(
    table: ResultsTable,
    dest_path: str,
    **kwargs: Any,
) -> None:
    import duckdb

    assert table and table.backend

    array = f"{', ARRAY TRUE' if kwargs.get('array') else ''}"
    compression = f", COMPRESSION {kwargs.get('compression')}"
    date_format = (
        f", DATEFORMAT '{kwargs.get('''date_format''')}'"
        if kwargs.get("date_format")
        else ""
    )
    ts_format = (
        f", TIMESTAMPFORMAT '{kwargs.get('''options.timestamp_format''')}'"
        if kwargs.get("options.timestamp_format")
        else ""
    )
    try:
        __export_table = table.backend.source_data
        duckdb.execute(
            f"copy (select * from __export_table) to '{dest_path}' "
            "(FORMAT JSON"
            f"{array}{compression}{date_format}{ts_format}"
            ")"
        )
    except (duckdb.Error, OSError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("DuckDB raised an error when writing your query to a JSON file."),
        ) from e


def _export_orc(
    table: ResultsTable,
    dest_path: str,
    batch_size: int | str = 1024,
    stripe_size: int | str = 67108864,
    compression_block_size: int | str = 65536,
    row_index_stride: int | str = 10000,
    padding_tolerance: float | str = 0.0,
    dictionary_key_size_threshold: float | str = 0.0,
    bloom_filter_columns: list[str] | str | None = None,
    bloom_filter_fpp: float | str = 0.05,
    **kwargs: Any,
) -> None:
    import pyarrow.lib as pl
    import pyarrow.orc as po

    assert table and table.backend

    try:
        if bloom_filter_columns and isinstance(bloom_filter_columns, str):
            bloom_filter_columns = bloom_filter_columns.split(",")
        batch_size = int(batch_size)
        compression_block_size = int(compression_block_size)
        stripe_size = int(stripe_size)
        row_index_stride = int(row_index_stride)
        bloom_filter_fpp = float(bloom_filter_fpp)
        padding_tolerance = float(padding_tolerance)
        dictionary_key_size_threshold = float(dictionary_key_size_threshold)
    except (ValueError, TypeError, KeyError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("Arrow raised an error when writing your data to an ORC file."),
        ) from e
    try:
        po.write_table(
            table.backend.source_data,
            dest_path,
            batch_size=batch_size,
            compression_block_size=compression_block_size,
            stripe_size=stripe_size,
            row_index_stride=row_index_stride,
            bloom_filter_fpp=bloom_filter_fpp,
            padding_tolerance=padding_tolerance,
            dictionary_key_size_threshold=dictionary_key_size_threshold,
            bloom_filter_columns=bloom_filter_columns,  # type: ignore
            **kwargs,
        )
    except (pl.ArrowException, OSError, IOError, TypeError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("Arrow raised an error when writing your data to an ORC file."),
        ) from e


def _export_feather(
    table: ResultsTable,
    dest_path: str,
    compression: str | None = None,
    compression_level: str | int | None = None,
    chunksize: str | int | None = None,
    **kwargs: Any,
) -> None:
    import pyarrow.feather as pf
    import pyarrow.lib as pl

    assert table and table.backend

    try:
        compression_level = int(compression_level) if compression_level else None
        chunksize = int(chunksize) if chunksize else None
    except (ValueError, TypeError, KeyError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("Arrow raised an error when writing your data to a Feather file."),
        ) from e

    try:
        pf.write_feather(
            table.backend.source_data,
            dest_path,
            compression=compression,
            compression_level=compression_level,
            chunksize=chunksize,
            **kwargs,
        )
    except (pl.ArrowException, OSError, IOError, TypeError) as e:
        raise HarlequinCopyError(
            str(e),
            title=("Arrow raised an error when writing your data to a Feather file."),
        ) from e



from pathlib import Path

from textual import events
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Markdown, Static


class VerticalSuppressClicks(Vertical):
    def on_click(self, message: events.Click) -> None:
        message.stop()


class HelpScreen(ModalScreen):
    header_text = """
        Welcome to Harlequin! This screen contains a small subset of the online
        docs, available at https://harlequin.sh/docs/getting-started
    """.split()

    def compose(self) -> ComposeResult:
        markdown_path = Path(__file__).parent / "help_screen.md"
        with open(markdown_path, "r") as f:
            markdown = f.read()

        with VerticalSuppressClicks(id="help_outer"):
            yield Static(" ".join(self.header_text), id="help_header")
            with VerticalScroll(id="help_inner"):
                yield Markdown(markdown=markdown)
            yield Static(
                "Scroll with arrows. Press any other key to continue.", id="help_footer"
            )

    def on_mount(self) -> None:
        container = self.query_one("#help_outer")
        container.border_title = "Harlequin Help"
        self.body = self.query_one("#help_inner")

    def on_key(self, event: events.Key) -> None:
        event.stop()
        if event.key == "up":
            self.body.scroll_up()
        elif event.key == "down":
            self.body.scroll_down()
        elif event.key == "left":
            self.body.scroll_left()
        elif event.key == "right":
            self.body.scroll_right()
        elif event.key == "pageup":
            self.body.scroll_page_up()
        elif event.key == "pagedown":
            self.body.scroll_page_down()
        else:
            self.app.pop_screen()

    def on_click(self) -> None:
        self.app.pop_screen()



from __future__ import annotations

from typing import ClassVar

from textual.widgets import (
    DirectoryTree,
)

from harlequin.components.data_catalog.tree import HarlequinTree as HarlequinTree
from harlequin.messages import WidgetMounted


class FileTree(HarlequinTree, DirectoryTree, inherit_bindings=False):
    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "directory-tree--extension",
        "directory-tree--file",
        "directory-tree--folder",
        "directory-tree--hidden",
    }

    def on_mount(self) -> None:
        self.guide_depth = 3
        self.root.expand()
        self.post_message(WidgetMounted(widget=self))



from __future__ import annotations

from typing import ClassVar, Generic, TypeVar, Union

from textual.events import Click
from textual.message import Message
from textual.widgets import (
    Tree,
)
from textual.widgets._directory_tree import DirEntry
from textual.widgets._tree import EventTreeDataType, TreeNode

from harlequin.catalog import CatalogItem

TTreeNode = TypeVar("TTreeNode")


class HarlequinTree(Tree[TTreeNode], inherit_bindings=False):
    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "harlequin-tree--type-label",
    }
    double_click: int | None = None

    class CatalogError(Message):
        def __init__(self, catalog_type: str, error: BaseException) -> None:
            self.catalog_type = catalog_type
            self.error = error
            super().__init__()

    class NodeSubmitted(Generic[EventTreeDataType], Message):
        def __init__(self, node: TreeNode[EventTreeDataType]) -> None:
            self.node: TreeNode[EventTreeDataType] = node
            super().__init__()

        @property
        def insert_name(self) -> str:
            if not self.node.data:
                return ""
            elif isinstance(self.node.data, CatalogItem):
                return self.node.data.query_name
            elif isinstance(self.node.data, DirEntry):
                return f"'{self.node.data.path}'"
            else:
                return str(self.node.data)

    class NodeCopied(Generic[EventTreeDataType], Message):
        def __init__(self, node: TreeNode[EventTreeDataType]) -> None:
            self.node: TreeNode[EventTreeDataType] = node
            super().__init__()

        @property
        def copy_name(self) -> str:
            if not self.node.data:
                return ""
            elif isinstance(self.node.data, CatalogItem):
                return self.node.data.query_name
            elif isinstance(self.node.data, DirEntry):
                return str(self.node.data.path)
            else:
                return str(self.node.data)

    class ShowContextMenu(Message):
        def __init__(self, node: TreeNode) -> None:
            self.node = node
            super().__init__()

    class HideContextMenu(Message):
        pass

    def on_focus(self) -> None:
        if self.cursor_line < 0:
            self.cursor_line = 0

    async def on_click(self, event: Click) -> None:
        meta = event.style.meta
        click_line: Union[int, None] = meta.get("line", None)
        if event.button == 1:  # left button click
            self.post_message(self.HideContextMenu())
            if (
                self.double_click is not None
                and click_line is not None
                and self.double_click == click_line
            ):
                event.prevent_default()
                node = self.get_node_at_line(click_line)
                if node is not None:
                    self.post_message(self.NodeSubmitted(node=node))
                    node.expand()
            else:
                self.double_click = click_line
                self.set_timer(
                    delay=0.5,
                    callback=self._clear_double_click,
                    name="double_click_timer",
                )
        elif event.button == 3 and click_line is not None:  # right click
            node = self.get_node_at_line(click_line)
            if node is not None and isinstance(node.data, CatalogItem):
                self.post_message(self.ShowContextMenu(node=node))

    def _clear_double_click(self) -> None:
        self.double_click = None

    def action_submit(self) -> None:
        if self.cursor_node is not None:
            self.post_message(self.NodeSubmitted(node=self.cursor_node))

    def action_copy(self) -> None:
        if self.cursor_node is not None:
            self.post_message(self.NodeCopied(node=self.cursor_node))

    def action_show_context_menu(self) -> None:
        if self.cursor_node is not None and isinstance(
            self.cursor_node.data, CatalogItem
        ):
            self.post_message(self.ShowContextMenu(node=self.cursor_node))

    def action_hide_context_menu(self) -> None:
        self.post_message(self.HideContextMenu())



from __future__ import annotations

import hashlib
import json
import pickle
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Sequence

from platformdirs import user_cache_dir

from harlequin.catalog import Catalog
from harlequin.history import History

if TYPE_CHECKING:
    from harlequin.components.data_catalog import S3Tree

CACHE_VERSION = 2


def recursive_dict() -> defaultdict:
    return defaultdict(recursive_dict)


class PermissiveEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, Path):
            return str(obj)
        # Never raise a TypeError, just use the repr
        try:
            return str(obj)
        except TypeError:
            return ""


@dataclass
class CatalogCache:
    databases: dict[str, Catalog]
    s3: dict[tuple[str | None, str | None, str | None], dict]
    history: dict[str, History]

    def get_db(self, connection_hash: str) -> Catalog | None:
        # if connection_hash:
        #     return self.databases.get(connection_hash, None)
        return None

    def get_history(self, connection_hash: str) -> History | None:
        if connection_hash:
            return self.history.get(connection_hash, None)
        return None

    def get_s3(
        self, cache_key: tuple[str | None, str | None, str | None]
    ) -> dict | None:
        return self.s3.get(cache_key, None)


def get_connection_hash(conn_str: Sequence[str], config: Mapping[str, Any]) -> str:
    return (
        hashlib.md5(
            json.dumps(
                {"conn_str": tuple(conn_str), **config},
                cls=PermissiveEncoder,
            ).encode("utf-8")
        )
        .digest()
        .hex()
    )


def get_catalog_cache() -> CatalogCache | None:
    return _load_cache()


def update_catalog_cache(
    connection_hash: str | None,
    catalog: Catalog | None,
    s3_tree: S3Tree | None,
    history: History | None,
) -> None:
    if connection_hash is None and s3_tree is None:
        return
    cache = _load_cache()
    if cache is None:
        cache = CatalogCache(databases={}, s3={}, history={})
    # if catalog is not None and connection_hash:
    #     cache.databases[connection_hash] = catalog
    if s3_tree is not None and s3_tree.catalog_data is not None:
        cache.s3[s3_tree.cache_key] = s3_tree.catalog_data
    if history is not None and connection_hash:
        cache.history[connection_hash] = history
    _write_cache(cache)


def _get_cache_file() -> Path:
    """
    Returns the path to the cache file on disk
    """
    cache_dir = Path(user_cache_dir(appname="harlequin"))
    cache_file = cache_dir / f"catalog-cache-{CACHE_VERSION}.pickle"
    return cache_file


def _load_cache() -> CatalogCache | None:
    """
    Returns a Cache by loading from a pickle saved to disk
    """
    cache_file = _get_cache_file()
    try:
        with cache_file.open("rb") as f:
            cache: CatalogCache = pickle.load(f)
            assert isinstance(cache, CatalogCache)
    except (
        pickle.UnpicklingError,
        ValueError,
        IndexError,
        FileNotFoundError,
        AssertionError,
        EOFError,
    ):
        return None
    else:
        return cache


def _write_cache(cache: CatalogCache) -> None:
    """
    Updates cache with current data catalog
    """
    cache_file = _get_cache_file()
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    with open(cache_file, "wb") as f:
        pickle.dump(cache, f)



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property


@dataclass(order=False)
class HarlequinCompletion:
    """
    A HarlequinCompletion represents a single item in the autocomplete drop-down list.

    Harlequin will concatenate and style the label and type_label and show them in
    the autocomplete list. When a user selects an autocomplete item, Harlequin will
    insert the completion's value into the editor. Completions are ordered by their
    priority (lowest first). Harlequin gives reserved keywords a priority of 100,
    Catalog items a priority of 500, and functions a priority of 1000. Completions
    can have a context, which will cause them to be shown as member completions after
    the user types the context followed by a `.` or `:`.

    Args:
        label (str): The text shown to the user in the autocomplete menu.
        type_label (str): Dimmend text shown to the user next to the label in the
            autocomplete menu.
        value (str): The text inserted into the query editor when the completion is
            selected.
        priority (int): The sort order of the completion (lowest first). Harlequin
            uses 100 for reserved keywords and 1000 for functions.
        context (str | None): A namespace for the completion. Defaults to None.
    """

    label: str
    type_label: str
    value: str
    priority: int
    context: str | None = None

    def __lt__(self, other: "HarlequinCompletion") -> bool:
        return (self.priority, self.label) < (other.priority, other.label)

    def __le__(self, other: "HarlequinCompletion") -> bool:
        return (self.priority, self.label) <= (other.priority, other.label)

    def __gt__(self, other: "HarlequinCompletion") -> bool:
        return (self.priority, self.label) > (other.priority, other.label)

    def __ge__(self, other: "HarlequinCompletion") -> bool:
        return (self.priority, self.label) >= (other.priority, other.label)

    @cached_property
    def match_val(self) -> str:
        return self.label.lower()



from __future__ import annotations

from typing import ClassVar

from rich.style import Style
from rich.text import Text
from textual.css.query import NoMatches
from textual.widgets import (
    ContentSwitcher,
    TabbedContent,
    TabPane,
    Tabs,
)
from textual_fastdatatable import DataTable
from textual_fastdatatable.backend import AutoBackendType

from harlequin.messages import WidgetMounted


class ResultsTable(DataTable, inherit_bindings=False):
    DEFAULT_CSS = """
        ResultsTable {
            height: 100%;
            width: 100%;
        }
    """

    def on_mount(self) -> None:
        self.post_message(WidgetMounted(widget=self))


class ResultsViewer(TabbedContent, can_focus=True):
    BORDER_TITLE = "Query Results"
    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "results-viewer--type-label",
    }

    def __init__(
        self,
        max_results: int = 10_000,
    ) -> None:
        super().__init__()
        self.max_results = max_results

    def on_mount(self) -> None:
        self.query_one(Tabs).can_focus = False
        self.add_class("hide-tabs")
        self.max_col_width = self._get_max_col_width()
        self.post_message(WidgetMounted(widget=self))

    def clear_all_tables(self) -> None:
        self.clear_panes()
        self.add_class("hide-tabs")

    def get_visible_table(self) -> ResultsTable | None:
        content = self.query_one(ContentSwitcher)
        active_tab_id = self.active
        if active_tab_id:
            try:
                tab_pane = content.query_one(f"#{active_tab_id}", TabPane)
                return tab_pane.query_one(ResultsTable)
            except NoMatches:
                return None
        else:
            tables = content.query(ResultsTable)
            try:
                return tables.first(ResultsTable)
            except NoMatches:
                return None

    async def push_table(
        self,
        table_id: str,
        column_labels: list[tuple[str, str]],
        data: AutoBackendType,
    ) -> ResultsTable:
        formatted_labels = [
            self._format_column_label(col_name, col_type)
            for col_name, col_type in column_labels
        ]
        table = ResultsTable(
            id=table_id,
            column_labels=formatted_labels,  # type: ignore
            data=data,
            max_rows=self.max_results,
            cursor_type="range",
            max_column_content_width=self.max_col_width,
            null_rep="[dim]∅ null[/]",
            render_markup=False,
        )
        n = self.tab_count + 1
        if n > 1:
            self.remove_class("hide-tabs")
        pane = TabPane(f"Result {n}", table, id=f"result-{n}")
        await self.add_pane(pane)
        # need to manually refresh the table, since activating the tab
        # doesn't consistently cause a new layout calc.
        table.refresh(repaint=True, layout=True)
        return table

    def show_loading(self) -> None:
        self.border_title = "Running Query"
        self.add_class("non-responsive")
        self.loading = True
        self.clear_all_tables()

    def show_table(self, did_run: bool = True) -> None:
        self.loading = False
        self.remove_class("non-responsive")
        if not did_run:
            self.border_title = "Query Results"
        else:
            table = self.get_visible_table()
            if table is not None:
                rows = table.source_row_count
                if rows > 0:
                    self.border_title = (
                        f"Query Results {self._human_row_count(table.source_row_count)}"
                    )
                else:
                    self.border_title = "Query Returned No Records"
            else:
                self.border_title = "Query Results"

    def on_focus(self) -> None:
        self._focus_on_visible_table()

    def on_resize(self) -> None:
        # only impacts new tables pushed after the resize
        self.max_col_width = self._get_max_col_width()

    def on_tabbed_content_tab_activated(
        self, message: TabbedContent.TabActivated
    ) -> None:
        message.stop()
        maybe_table = self.get_visible_table()
        if maybe_table is not None:
            self.border_title = (
                f"Query Results {self._human_row_count(maybe_table.source_row_count)}"
            )
            maybe_table.focus()

    def action_switch_tab(self, offset: int) -> None:
        if not self.active:
            return
        name_prefix, _, tab_number_str = self.active.rpartition("-")
        tab_number = int(tab_number_str)
        unsafe_tab_number = tab_number + offset
        if unsafe_tab_number < 1:
            new_tab_number = self.tab_count
        elif unsafe_tab_number > self.tab_count:
            new_tab_number = 1
        else:
            new_tab_number = unsafe_tab_number
        self.active = f"{name_prefix}-{new_tab_number}"
        self._focus_on_visible_table()

    def action_focus_data_catalog(self) -> None:
        if hasattr(self.app, "action_focus_data_catalog"):
            self.app.action_focus_data_catalog()

    def action_focus_query_editor(self) -> None:
        if hasattr(self.app, "action_focus_query_editor"):
            self.app.action_focus_query_editor()

    def _focus_on_visible_table(self) -> None:
        maybe_table = self.get_visible_table()
        if maybe_table is not None:
            maybe_table.focus()

    def _human_row_count(self, total_rows: int) -> str:
        if self.max_results > 0 and total_rows > self.max_results:
            return f"(Showing {self.max_results:,} of {total_rows:,} Records)"
        else:
            return f"({total_rows:,} Records)"

    def _format_column_label(self, col_name: str, col_type: str) -> Text:
        type_label_style = self.get_component_rich_style("results-viewer--type-label")
        type_label_fg_style = Style(color=type_label_style.color)
        label = Text.assemble(col_name, " ", (col_type, type_label_fg_style))
        return label

    def _get_max_col_width(self) -> int:
        SMALLEST_MAX_WIDTH = 20
        CELL_X_PADDING = 2
        parent_size = getattr(self.parent, "container_size", self.screen.container_size)
        return max(SMALLEST_MAX_WIDTH, parent_size.width // 2 - CELL_X_PADDING)



from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ConfirmModal(ModalScreen[bool]):
    def __init__(self, prompt: str) -> None:
        super().__init__()
        self.prompt = prompt

    def compose(self) -> ComposeResult:
        with Vertical(id="outer"):
            yield Label(self.prompt, id="prompt_label")
            with Horizontal(id="button_row"):
                yield Button(label="No", id="no")
                yield Button(label="Yes", variant="primary", id="yes")

    @on(Button.Pressed, "#yes")
    def save_from_button(self) -> None:
        self.action_continue()

    @on(Button.Pressed, "#no")
    def cancel_from_button(self) -> None:
        self.action_cancel()

    def action_continue(self) -> None:
        self.dismiss(True)

    def action_cancel(self) -> None:
        self.dismiss(False)



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Generator, Iterable, Sequence

import click
import questionary
from textual.validation import ValidationResult, Validator
from textual.widget import Widget

from harlequin.colors import HARLEQUIN_QUESTIONARY_STYLE
from harlequin.copy_widgets import (
    Input,
    NoFocusLabel,
    PathInput,
    Select,
    Switch,
)


class _CustomValidator(Validator):
    def __init__(
        self,
        validator: Callable[[str], tuple[bool, str | None]] | None = None,
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description)
        self.validator = validator or (lambda _: (True, ""))

    def validate(self, value: str) -> ValidationResult:
        try:
            is_valid, message = self.validator(value)
        except Exception as e:
            return self.failure(str(e))

        if is_valid:
            return self.success()
        else:
            return self.failure(message or "Validation failed.")


def concatenate(first: str, second: str) -> str:
    if first == second:
        return first
    return f"{first}\n----or----\n{second}"


class AbstractOption(ABC):
    """
    The ABC for Harlequin options that are used as both command-line options and
    GUI options. Options have names and descriptions, and may have user-facing
    labels and aliased, short declarations (for CLI options).

    Subclasses define options for specific data types, like text or boolean options.
    """

    def __init__(
        self,
        name: str,
        description: str,
        *args: Any,
        label: str | None = None,
        short_decls: Sequence[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Args:
            name (str): A unique name for this option. Must be a valid
                HTML/CSS id and a valid CLI option name (without the `--` prefix).
                e.g., "port", "header"
            description (str): Help text for this option.
            label (str | None): For GUI options, a human-friendly label for this option.
            short_decls (Sequence[str] | None): For CLI options, a list of short aliases
                (including the `-` prefix) for this option (e.g., ["-p"]).
        """
        # names should be valid html/css ids
        if re.match(r"[A-Za-z](\w|-)*", name):
            self.name = name
        else:
            raise ValueError(
                "An Option's name attribute must match "
                r"""r'[A-Za-z](\w|-)*' """
                "so it is a valid CLI flag and HTML/CSS id."
            )
        self.description = description
        self.label = label or name.replace("_", " ").replace("_", " ").capitalize()
        short_decls = short_decls or []
        self.short_decls = [
            decl if decl.startswith("-") else f"-{decl}" for decl in short_decls
        ]

    @abstractmethod
    def merge(self, other: AbstractOption) -> AbstractOption:
        """
        Merges two options together; used for options with the same name, to return
        a concatenated description and other merged properties.
        """
        pass

    @abstractmethod
    def to_click(self) -> Callable[[click.Command], click.Command]:
        pass

    @abstractmethod
    def to_widgets(self) -> Generator[Widget, None, None]:
        pass

    @abstractmethod
    def to_questionary(self, existing_value: Any | None = None) -> questionary.Question:
        pass


class TextOption(AbstractOption):
    """
    An option for free text input, including optional validation.
    """

    def __init__(
        self,
        name: str,
        description: str,
        label: str | None = None,
        short_decls: list[str] | None = None,
        default: str | None = None,
        placeholder: str | None = None,
        validator: Callable[[str], tuple[bool, str | None]] | None = None,
    ) -> None:
        """
        Args:
            name (str): A unique name for this option. Must be a valid
                HTML/CSS id and a valid CLI option name (without the `--` prefix).
                e.g., "port", "header"
            description (str): Help text for this option.
            label (str | None): For GUI options, a human-friendly label for this option.
            short_decls (Sequence[str] | None): For CLI options, a list of short aliases
                (including the `-` prefix) for this option (e.g., ["-p"]).
            default (str | None): The default value for this option.
            placeholder (str | None): For GUI options, placeholder text for this option.
            validator (Callable[[str], tuple[bool, str | None]] | None): A callable that
                receives the raw input as a string returns a tuple. The first item of
                the tuple is either True for valid input or False for invalid input.
                The second item is a message shown to the user if the validation fails.
        """
        super().__init__(name, description, label=label, short_decls=short_decls)
        self.validator = validator
        self.default = default
        self.placeholder = placeholder

    def merge(self, other: AbstractOption) -> AbstractOption:
        if isinstance(other, ListOption):
            return other.merge(self)

        name = self.name
        description = concatenate(self.description, other.description)
        label = self.label or other.label
        short_decls = set(self.short_decls) | set(other.short_decls)
        default = (
            self.default if self.default == getattr(other, "default", None) else None
        )
        placeholder = self.placeholder or getattr(other, "placeholder", None)

        def merge_validator(raw: str) -> tuple[bool, str | None]:
            """
            The merged validator must return true if either validator
            accepts the input; if the other Option does not have
            a validator, it accepts all inputs, so the merged validator
            must also.
            """
            if (
                self.validator is not None
                and isinstance(other, TextOption)
                and other.validator is not None
            ):
                result = self.validator(raw)
                if result[0]:
                    return result
                else:
                    return other.validator(raw)
            else:
                return True, None

        return TextOption(
            name=name,
            description=description,
            label=label,
            short_decls=list(short_decls),
            default=default,
            placeholder=placeholder,
            validator=merge_validator if self.validator is not None else None,
        )

    def to_click(self) -> Callable[[click.Command], click.Command]:
        def click_callback(
            ctx: click.Context, param: click.ParamType, value: str
        ) -> str:
            if self.validator is not None:
                try:
                    is_valid, message = self.validator(value)
                except Exception as e:
                    raise click.BadParameter(str(e)) from e
                if not is_valid:
                    raise click.BadParameter(message or "Validation failed.")
            return value

        return click.option(
            f"--{self.name}",
            *self.short_decls,
            help=self.description,
            callback=click_callback,
        )

    def to_widgets(self) -> Generator[Widget, None, None]:
        yield NoFocusLabel(f"{self.label}:", classes="input_label")
        yield Input(
            value=self.default or "",
            placeholder=self.placeholder or "",
            id=self.name,
            validators=[_CustomValidator(self.validator)],
        )

    def to_questionary(self, existing_value: Any | None = None) -> questionary.Question:
        def _q_validator(raw: str) -> bool | str | None:
            if self.validator is not None:
                result = self.validator(raw)
                if result[0]:
                    return True
                else:
                    return result[1]
            else:
                return True

        try:
            safe_existing_value = str(existing_value)
        except (ValueError, TypeError):
            safe_existing_value = None

        return questionary.text(
            message=self.name,
            default=(
                safe_existing_value
                if safe_existing_value is not None
                else self.default or ""
            ),
            validate=_q_validator,
            style=HARLEQUIN_QUESTIONARY_STYLE,
        )


class ListOption(AbstractOption):
    def __init__(
        self,
        name: str,
        description: str,
        label: str | None = None,
        short_decls: list[str] | None = None,
    ) -> None:
        """
        Args:
            name (str): A unique name for this option. Must be a valid
                HTML/CSS id and a valid CLI option name (without the `--` prefix).
                e.g., "port", "header"
            description (str): Help text for this option.
            label (str | None): For GUI options, a human-friendly label for this option.
            short_decls (Sequence[str] | None): For CLI options, a list of short aliases
                (including the `-` prefix) for this option (e.g., ["-p"]).
        """
        super().__init__(name, description, label=label, short_decls=short_decls)

    def merge(self, other: AbstractOption) -> ListOption:
        name = self.name
        description = concatenate(self.description, other.description)
        label = self.label or other.label
        short_decls = set(self.short_decls) | set(other.short_decls)
        return ListOption(
            name=name,
            description=description,
            label=label,
            short_decls=list(short_decls),
        )

    def to_click(self) -> Callable[[click.Command], click.Command]:
        return click.option(
            f"--{self.name}",
            *self.short_decls,
            help=self.description,
            multiple=True,
        )

    def to_widgets(self) -> Generator[Widget, None, None]:
        raise NotImplementedError("No widget for ListOption.")

    def to_questionary(self, existing_value: Any | None = None) -> questionary.Question:
        if isinstance(existing_value, str):
            safe_existing_value = existing_value
        elif isinstance(existing_value, Iterable):
            safe_existing_value = " ".join(existing_value)
        else:
            safe_existing_value = None

        return questionary.text(
            message=self.name,
            instruction="Separate items by a space.",
            default=safe_existing_value if safe_existing_value is not None else "",
            style=HARLEQUIN_QUESTIONARY_STYLE,
        )


class PathOption(AbstractOption):
    """
    A text input with path validation and autocomplete features.
    """

    def __init__(
        self,
        name: str,
        description: str,
        label: str | None = None,
        short_decls: list[str] | None = None,
        exists: bool = False,
        file_okay: bool = True,
        dir_okay: bool = True,
        resolve_path: bool = False,
        path_type: type | None = Path,
        default: str | None = None,
        placeholder: str | None = None,
    ) -> None:
        """
        Args:
            name (str): A unique name for this option. Must be a valid
                HTML/CSS id and a valid CLI option name (without the `--` prefix).
                e.g., "port", "header"
            description (str): Help text for this option.
            label (str | None): For GUI options, a human-friendly label for this option.
            short_decls (Sequence[str] | None): For CLI options, a list of short aliases
                (including the `-` prefix) for this option (e.g., ["-p"]).
            exists (bool): *Validation* Set True if the path must already exist.
            file_okay (bool): *Validation* Set True if the path may be a file.
            dir_okay (bool):  *Validation* Set True if the path may be a directory.
            resolve_path (bool): For CLI Options, set True for the returned path to be
                resolved.
            path_type (type): For CLI Options, define a type for the returned path
                (usually str or pathlib.Path).
            default (str): The default path.
            placeholder (str): For GUI options, the placeholder text for the input.
        """
        super().__init__(name, description, label=label, short_decls=short_decls)
        self.exists = exists
        self.file_okay = file_okay
        self.dir_okay = dir_okay
        self.resolve_path = resolve_path
        self.path_type = path_type
        self.default = default
        self.placeholder = placeholder

    def merge(self, other: AbstractOption) -> AbstractOption:
        if isinstance(other, (TextOption, ListOption)):
            return other.merge(self)
        name = self.name
        description = concatenate(self.description, other.description)
        label = self.label or other.label
        short_decls = set(self.short_decls) | set(other.short_decls)
        default = (
            self.default if self.default == getattr(other, "default", None) else None
        )
        placeholder = self.placeholder or getattr(other, "placeholder", None)
        if isinstance(other, PathOption):
            exists = self.exists and other.exists
            file_okay = self.file_okay or other.file_okay
            dir_okay = self.dir_okay or other.dir_okay
            resolve_path = self.resolve_path or other.resolve_path
            path_type = self.path_type if self.path_type == other.path_type else str
        else:
            exists = False
            file_okay = True
            dir_okay = True
            resolve_path = False
            path_type = str
        return PathOption(
            name=name,
            description=description,
            label=label,
            short_decls=list(short_decls),
            exists=exists,
            file_okay=file_okay,
            dir_okay=dir_okay,
            resolve_path=resolve_path,
            path_type=path_type,
            default=default,
            placeholder=placeholder,
        )

    def to_click(self) -> Callable[[click.Command], click.Command]:
        return click.option(
            f"--{self.name}",
            *self.short_decls,
            help=self.description,
            type=click.Path(
                exists=self.exists,
                file_okay=self.file_okay,
                dir_okay=self.dir_okay,
                resolve_path=self.resolve_path,
                path_type=self.path_type,
            ),
        )

    def to_widgets(self) -> Generator[Widget, None, None]:
        yield NoFocusLabel(f"{self.label}:", classes="input_label")
        yield PathInput(
            value=self.default or "",
            placeholder=self.placeholder or "",
            id=self.name,
            file_okay=self.file_okay,
            dir_okay=self.dir_okay,
            must_exist=self.exists,
            tab_advances_focus=True,
        )

    def to_questionary(self, existing_value: Any | None = None) -> questionary.Question:
        def _path_validator(raw_path: str) -> bool | str:
            try:
                p = Path(raw_path)
            except ValueError as e:
                return f"Not a valid path! {e}"
            if self.exists and not p.exists():
                return f"No file exists at {p}"

            if not self.file_okay and p.is_file():
                return f"{p} is a file, expected a directory."

            if not self.dir_okay and p.is_dir():
                return f"{p} is a directory, expected a file."

            return True

        try:
            safe_existing_value = str(existing_value)
        except (ValueError, TypeError):
            safe_existing_value = None

        return questionary.path(
            message=self.name,
            default=(
                safe_existing_value
                if safe_existing_value is not None
                else self.default or ""
            ),
            only_directories=not self.file_okay,
            validate=_path_validator,
            style=HARLEQUIN_QUESTIONARY_STYLE,
        )


class SelectOption(AbstractOption):
    def __init__(
        self,
        name: str,
        description: str,
        choices: Sequence[str | tuple[str, str]],
        label: str | None = None,
        short_decls: list[str] | None = None,
        default: str | None = None,
    ) -> None:
        super().__init__(name, description, label=label, short_decls=short_decls)
        self.choices = choices
        self.default = default
        """
        Args:
            name (str): A unique name for this option. Must be a valid
                HTML/CSS id and a valid CLI option name (without the `--` prefix).
                e.g., "port", "header"
            description (str): Help text for this option.
            choices (Sequence[str | tuple[str, str]]): A list of values or list of
                (label, value) pairs for the user to select from.
            label (str | None): For GUI options, a human-friendly label for this option.
            short_decls (Sequence[str] | None): For CLI options, a list of short aliases
                (including the `-` prefix) for this option (e.g., ["-p"]).
            default (str | None): The default value for this option.
        """

    def merge(self, other: AbstractOption) -> AbstractOption:
        if isinstance(other, (TextOption, PathOption, ListOption)):
            return other.merge(self)
        name = self.name
        description = concatenate(self.description, other.description)
        label = self.label or other.label
        short_decls = set(self.short_decls) | set(other.short_decls)
        choices = set(self.choices) | set(getattr(other, "choices", []))
        default = (
            self.default if self.default == getattr(other, "default", None) else None
        )
        return SelectOption(
            name=name,
            description=description,
            choices=list(choices),
            label=label,
            short_decls=list(short_decls),
            default=default,
        )

    def to_click(self) -> Callable[[click.Command], click.Command]:
        return click.option(
            f"--{self.name}",
            *self.short_decls,
            help=self.description,
            type=click.Choice(choices=self._flat_choices(), case_sensitive=False),
        )

    def to_widgets(self) -> Generator[Widget, None, None]:
        choices: list[tuple[str, str]] = []
        for choice in self.choices:
            if isinstance(choice, str):
                choices.append((choice, choice))
            else:
                choices.append(choice)
        yield NoFocusLabel(f"{self.label}:", classes="select_label")
        yield Select(
            options=choices,
            id=self.name,
            value=self.default,
            allow_blank=False,
        )

    def to_questionary(self, existing_value: Any | None = None) -> questionary.Question:
        try:
            safe_existing_value = str(existing_value)
        except (ValueError, TypeError):
            safe_existing_value = None

        if safe_existing_value not in self._flat_choices():
            safe_existing_value = None

        return questionary.select(
            message=self.name,
            choices=self._flat_choices(),
            default=(
                safe_existing_value if safe_existing_value is not None else self.default
            ),
            style=HARLEQUIN_QUESTIONARY_STYLE,
        )

    def _flat_choices(self) -> list[str]:
        choices: list[str] = []
        for choice in self.choices:
            if isinstance(choice, str):
                choices.append(choice)
            else:
                choices.append(choice[0])
        return choices


class FlagOption(AbstractOption):
    """
    A boolean option, defaults to False. (Can set another default, but that only applies
    for GUI options, not CLI options, which always default to False)
    """

    def __init__(
        self,
        name: str,
        description: str,
        label: str | None = None,
        short_decls: Sequence[str] | None = None,
        default: bool = False,
    ) -> None:
        super().__init__(name, description, label=label, short_decls=short_decls)
        self.default = default

    def merge(self, other: AbstractOption) -> AbstractOption:
        if not isinstance(other, FlagOption):
            return other.merge(self)
        name = self.name
        description = concatenate(self.description, other.description)
        label = self.label or other.label
        short_decls = set(self.short_decls) | set(other.short_decls)
        default = self.default and other.default
        return FlagOption(
            name=name,
            description=description,
            label=label,
            short_decls=list(short_decls),
            default=default,
        )

    def to_click(self) -> Callable[[click.Command], click.Command]:
        return click.option(
            f"--{self.name}", *self.short_decls, help=self.description, is_flag=True
        )

    def to_widgets(self) -> Generator[Widget, None, None]:
        yield NoFocusLabel(f"{self.label}:", classes="switch_label")
        yield Switch(value=self.default, id=self.name)

    def to_questionary(self, existing_value: Any | None = None) -> questionary.Question:
        try:
            safe_existing_value = bool(existing_value)
        except (ValueError, TypeError):
            safe_existing_value = None

        return questionary.confirm(
            message=self.name,
            default=safe_existing_value if safe_existing_value is not None else False,
            style=HARLEQUIN_QUESTIONARY_STYLE,
        )


HarlequinAdapterOption = AbstractOption


class HarlequinCopyFormat:
    """
    A file format for data export that is supported by Harlequin.
    """

    name: str
    label: str
    extensions: Sequence[str]
    options: Sequence[HarlequinAdapterOption]

    def __init__(
        self,
        name: str,
        label: str | None = None,
        extensions: Sequence[str] | None = None,
        options: Sequence[HarlequinAdapterOption] | None = None,
    ) -> None:
        """
        Args:
            name (str): A unique, internal name for this format. E.g., 'csv'
            label (str | None): A user-facing name for this format. E.g., 'CSV'
            extensions (Sequence[str] | None): A seq of file extensions to associate
                with this format. Should include the leading period. e.g.,
                (".csv", ".tsv").
            options (Sequence[HarlequinAdapterOption] | None): A list of options for
                configuring copy operations for this format. e.g.,
                [FlagOption(name="Header", description="Include header row?")]
        """
        self.name = name
        self.label = label or name.replace("_", " ").replace("_", " ").capitalize()
        self.extensions = extensions or tuple()
        self.extensions = [
            ext if ext.startswith(".") else f".{ext}" for ext in self.extensions
        ]
        self.options = options or tuple()



from harlequin.keymap import HarlequinKeyBinding, HarlequinKeyMap

VSCODE_APP_BINDINGS = [
    HarlequinKeyBinding("ctrl+q", "quit"),
    HarlequinKeyBinding("f1", "help"),
    HarlequinKeyBinding("f2", "focus_query_editor"),
    HarlequinKeyBinding("f5", "focus_results_viewer"),
    HarlequinKeyBinding("f6", "focus_data_catalog"),
    HarlequinKeyBinding("f8", "show_query_history"),
    HarlequinKeyBinding("ctrl+b,f9", "toggle_sidebar"),
    HarlequinKeyBinding("f10", "toggle_full_screen"),
    HarlequinKeyBinding("ctrl+e", "show_data_exporter"),
    HarlequinKeyBinding("ctrl+r", "refresh_catalog"),
    HarlequinKeyBinding("tab", "focus_next"),
    HarlequinKeyBinding("shift+tab", "focus_previous"),
]
VSCODE_EDITOR_BINDINGS = [
    HarlequinKeyBinding("ctrl+n", "code_editor.new_buffer"),
    HarlequinKeyBinding("ctrl+w", "code_editor.close_buffer"),
    HarlequinKeyBinding("ctrl+k", "code_editor.next_buffer"),
    HarlequinKeyBinding(
        "ctrl+enter,ctrl+j",
        "code_editor.run_query",
        key_display="^⏎ or ^j",
    ),
    HarlequinKeyBinding("f4", "code_editor.format_buffer"),
    HarlequinKeyBinding("ctrl+s", "code_editor.save_buffer"),
    HarlequinKeyBinding("ctrl+o", "code_editor.load_buffer"),
    HarlequinKeyBinding("ctrl+f", "code_editor.find"),
    HarlequinKeyBinding("f3", "code_editor.find_next"),
    HarlequinKeyBinding("ctrl+g", "code_editor.goto_line"),
    HarlequinKeyBinding("up", "code_editor.cursor_up"),
    HarlequinKeyBinding("down", "code_editor.cursor_down"),
    HarlequinKeyBinding("left", "code_editor.cursor_left"),
    HarlequinKeyBinding("right", "code_editor.cursor_right"),
    HarlequinKeyBinding("ctrl+left", "code_editor.cursor_word_left"),
    HarlequinKeyBinding("ctrl+right", "code_editor.cursor_word_right"),
    HarlequinKeyBinding("home", "code_editor.cursor_line_start"),
    HarlequinKeyBinding("end", "code_editor.cursor_line_end"),
    HarlequinKeyBinding("ctrl+home", "code_editor.cursor_doc_start"),
    HarlequinKeyBinding("ctrl+end", "code_editor.cursor_doc_end"),
    HarlequinKeyBinding("pageup", "code_editor.cursor_page_up"),
    HarlequinKeyBinding("pagedown", "code_editor.cursor_page_down"),
    HarlequinKeyBinding("shift+up", "code_editor.select_up"),
    HarlequinKeyBinding("shift+down", "code_editor.select_down"),
    HarlequinKeyBinding("shift+left", "code_editor.select_left"),
    HarlequinKeyBinding("shift+right", "code_editor.select_right"),
    HarlequinKeyBinding("ctrl+shift+left", "code_editor.select_word_left"),
    HarlequinKeyBinding("ctrl+shift+right", "code_editor.select_word_right"),
    HarlequinKeyBinding("shift+home", "code_editor.select_line_start"),
    HarlequinKeyBinding("shift+end", "code_editor.select_line_end"),
    HarlequinKeyBinding("ctrl+shift+home", "code_editor.select_doc_start"),
    HarlequinKeyBinding("ctrl+shift+end", "code_editor.select_doc_end"),
    HarlequinKeyBinding("ctrl+a", "code_editor.select_all"),
    HarlequinKeyBinding("ctrl+up", "code_editor.scroll_up_one"),
    HarlequinKeyBinding("ctrl+down", "code_editor.scroll_down_one"),
    HarlequinKeyBinding("ctrl+underscore", "code_editor.toggle_comment"),
    HarlequinKeyBinding("ctrl+x", "code_editor.cut"),
    HarlequinKeyBinding("ctrl+c", "code_editor.copy"),
    HarlequinKeyBinding("ctrl+u,ctrl+v,shift+insert", "code_editor.paste"),
    HarlequinKeyBinding("ctrl+z", "code_editor.undo"),
    HarlequinKeyBinding("ctrl+y", "code_editor.redo"),
    HarlequinKeyBinding("backspace", "code_editor.delete_left"),
    HarlequinKeyBinding("delete", "code_editor.delete_right"),
    HarlequinKeyBinding("shift+delete", "code_editor.delete_line"),
]
VSCODE_DATA_CATALOG_BINDINGS = [
    HarlequinKeyBinding("j", "data_catalog.previous_tab"),
    HarlequinKeyBinding("k", "data_catalog.next_tab"),
    HarlequinKeyBinding(
        "ctrl+enter,ctrl+j",
        "data_catalog.insert_name",
        key_display="^⏎ or ^j",
    ),
    HarlequinKeyBinding("ctrl+c", "data_catalog.copy_name"),
    HarlequinKeyBinding("enter", "data_catalog.select_cursor"),
    HarlequinKeyBinding("space", "data_catalog.toggle_node"),
    HarlequinKeyBinding("up", "data_catalog.cursor_up"),
    HarlequinKeyBinding("down", "data_catalog.cursor_down"),
    HarlequinKeyBinding("full_stop", "data_catalog.show_context_menu"),
    HarlequinKeyBinding("escape", "data_catalog.hide_context_menu"),
]
VSCODE_RESULTS_VIEWER_BINDINGS = [
    HarlequinKeyBinding("j", "results_viewer.previous_tab"),
    HarlequinKeyBinding("k", "results_viewer.next_tab"),
    HarlequinKeyBinding("ctrl+c", "results_viewer.copy_selection"),
    HarlequinKeyBinding("enter", "results_viewer.select_cursor"),
    HarlequinKeyBinding("up", "results_viewer.cursor_up"),
    HarlequinKeyBinding("down", "results_viewer.cursor_down"),
    HarlequinKeyBinding("left", "results_viewer.cursor_left"),
    HarlequinKeyBinding("right", "results_viewer.cursor_right"),
    HarlequinKeyBinding("ctrl+left", "results_viewer.cursor_row_start"),
    HarlequinKeyBinding("ctrl+right", "results_viewer.cursor_row_end"),
    HarlequinKeyBinding("ctrl+up,home", "results_viewer.cursor_column_start"),
    HarlequinKeyBinding("ctrl+down,end", "results_viewer.cursor_column_end"),
    HarlequinKeyBinding("tab", "results_viewer.cursor_next_cell"),
    HarlequinKeyBinding("shift+tab", "results_viewer.cursor_previous_cell"),
    HarlequinKeyBinding("pageup", "results_viewer.cursor_page_up"),
    HarlequinKeyBinding("pagedown", "results_viewer.cursor_page_down"),
    HarlequinKeyBinding("ctrl+home", "results_viewer.cursor_table_start"),
    HarlequinKeyBinding("ctrl+end", "results_viewer.cursor_table_end"),
    HarlequinKeyBinding("shift+up", "results_viewer.select_up"),
    HarlequinKeyBinding("shift+down", "results_viewer.select_down"),
    HarlequinKeyBinding("shift+left", "results_viewer.select_left"),
    HarlequinKeyBinding("shift+right", "results_viewer.select_right"),
    HarlequinKeyBinding("ctrl+shift+left", "results_viewer.select_row_start"),
    HarlequinKeyBinding("ctrl+shift+right", "results_viewer.select_row_end"),
    HarlequinKeyBinding(
        "ctrl+shift+up,shift+home", "results_viewer.select_column_start"
    ),
    HarlequinKeyBinding(
        "ctrl+shift+down,shift+end", "results_viewer.select_column_end"
    ),
    HarlequinKeyBinding("shift+pageup", "results_viewer.select_page_up"),
    HarlequinKeyBinding("shift+pagedown", "results_viewer.select_page_down"),
    HarlequinKeyBinding("ctrl+shift+home", "results_viewer.select_table_start"),
    HarlequinKeyBinding("ctrl+shift+end", "results_viewer.select_table_end"),
    HarlequinKeyBinding("ctrl+a", "results_viewer.select_all"),
]

VSCODE_HISTORY_SCREEN_BINDINGS = [
    HarlequinKeyBinding("enter", "history_screen.select_query"),
    HarlequinKeyBinding("escape", "history_screen.cancel"),
]


VSCODE = HarlequinKeyMap(
    name="vscode",
    bindings=[
        *VSCODE_APP_BINDINGS,
        *VSCODE_EDITOR_BINDINGS,
        *VSCODE_DATA_CATALOG_BINDINGS,
        *VSCODE_RESULTS_VIEWER_BINDINGS,
        *VSCODE_HISTORY_SCREEN_BINDINGS,
    ],
)



from pathlib import Path

from harlequin.options import (
    FlagOption,
    ListOption,
    PathOption,
    TextOption,
)

init = PathOption(
    name="init-path",
    description=(
        "The path to an initialization script. On startup, Harlequin will execute "
        "the commands in the script against the attached database."
    ),
    short_decls=["-i", "-init"],
    exists=False,
    file_okay=True,
    dir_okay=False,
    resolve_path=True,
    path_type=Path,
)

no_init = FlagOption(
    name="no-init",
    description="Start Harlequin without executing the initialization script.",
)

read_only = FlagOption(
    name="read-only",
    short_decls=["-readonly", "-r"],
    description="Open the database file in read-only mode.",
)

unsigned = FlagOption(
    name="allow-unsigned-extensions",
    description="Allow loading unsigned extensions",
    short_decls=["-u", "-unsigned"],
)

extensions = ListOption(
    name="extension",
    description=(
        "Install and load the named DuckDB extension when starting "
        "Harlequin. To install multiple extensions, repeat this option."
    ),
    short_decls=["-e"],
)

force_extensions = FlagOption(
    name="force-install-extensions",
    description="Force install all extensions passed with -e.",
)

custom_extension_repo = TextOption(
    name="custom-extension-repo",
    description=(
        "A value to pass to DuckDB's custom_extension_repository variable. "
        "Will be set before installing any extensions that are passed using -e."
    ),
)

md_token = TextOption(
    name="md_token",
    description=(
        "MotherDuck Token. Pass your MotherDuck service token in this option, or "
        "set the `motherduck_token` environment variable."
    ),
)

md_saas = FlagOption(
    name="md_saas",
    description="Run MotherDuck in SaaS mode (no local privileges).",
)

DUCKDB_OPTIONS = [
    init,
    no_init,
    read_only,
    unsigned,
    extensions,
    force_extensions,
    custom_extension_repo,
    md_token,
    md_saas,
]
